@@ -5,6 +5,7 @@
 #include "system.h"
 
 #include <netinet/in.h>
+#include <pthread.h>
 
 #include <rpm/rpmlib.h>			/* XXX RPMSIGTAG, other sig stuff */
 #include <rpm/rpmts.h>
@@ -20,14 +21,12 @@
 
 #include "debug.h"
 
-static int _print_pkts = 0;
-
-static const unsigned int nkeyids_max = 256;
-static unsigned int nkeyids = 0;
-static unsigned int nextkeyid  = 0;
-static unsigned int * keyids;
-
-void headerMergeLegacySigs(Header h, const Header sigh)
+/** \ingroup header
+ * Translate and merge legacy signature tags into header.
+ * @param h		header (dest)
+ * @param sigh		signature header (src)
+ */
+static void headerMergeLegacySigs(Header h, Header sigh)
 {
     HeaderIterator hi;
     struct rpmtd_s td;
@@ -94,77 +93,43 @@ void headerMergeLegacySigs(Header h, const Header sigh)
 	    (void) headerPut(h, &td, HEADERPUT_DEFAULT);
 	}
     }
-    hi = headerFreeIterator(hi);
+    headerFreeIterator(hi);
 }
 
-Header headerRegenSigHeader(const Header h, int noArchiveSize)
+static unsigned int getKeyid(pgpDigParams sigp)
 {
-    Header sigh = rpmNewSignature();
-    HeaderIterator hi;
-    struct rpmtd_s td;
-
-    for (hi = headerInitIterator(h); headerNext(hi, &td); rpmtdFreeData(&td)) {
-	switch (td.tag) {
-	/* XXX Translate legacy signature tag values. */
-	case RPMTAG_SIGSIZE:
-	    td.tag = RPMSIGTAG_SIZE;
-	    break;
-	case RPMTAG_SIGPGP:
-	    td.tag = RPMSIGTAG_PGP;
-	    break;
-	case RPMTAG_SIGMD5:
-	    td.tag = RPMSIGTAG_MD5;
-	    break;
-	case RPMTAG_SIGGPG:
-	    td.tag = RPMSIGTAG_GPG;
-	    break;
-	case RPMTAG_SIGPGP5:
-	    td.tag = RPMSIGTAG_PGP5;
-	    break;
-	case RPMTAG_ARCHIVESIZE:
-	    /* XXX rpm-4.1 and later has archive size in signature header. */
-	    if (noArchiveSize)
-		continue;
-	    td.tag = RPMSIGTAG_PAYLOADSIZE;
-	    break;
-	case RPMTAG_SHA1HEADER:
-	case RPMTAG_DSAHEADER:
-	case RPMTAG_RSAHEADER:
-	default:
-	    if (!(td.tag >= HEADER_SIGBASE && td.tag < HEADER_TAGBASE))
-		continue;
-	    break;
-	}
-	if (td.data == NULL) continue;	/* XXX can't happen */
-	if (!headerIsEntry(sigh, td.tag))
-	    (void) headerPut(sigh, &td, HEADERPUT_DEFAULT);
-    }
-    hi = headerFreeIterator(hi);
-    return sigh;
+    return (sigp != NULL) ? pgpGrab(sigp->signid+4, 4) : 0;
 }
 
 /**
  * Remember current key id.
- * @param dig		OpenPGP packet containter
+ * XXX: This s*** needs to die. Hook it into keyring or sumthin...
+ * @param keyid		signature keyid
  * @return		0 if new keyid, otherwise 1
  */
-static int stashKeyid(pgpDig dig)
+static int stashKeyid(unsigned int keyid)
 {
-    pgpDigParams sigp = dig ? &dig->signature : NULL;
-    unsigned int keyid;
-    int i;
+    static pthread_mutex_t keyid_lock = PTHREAD_MUTEX_INITIALIZER;
+    static const unsigned int nkeyids_max = 256;
+    static unsigned int nkeyids = 0;
+    static unsigned int nextkeyid  = 0;
+    static unsigned int * keyids;
 
-    if (dig == NULL || sigp == NULL)
+    int i;
+    int seen = 0;
+
+    if (keyid == 0)
 	return 0;
 
-    keyid = pgpGrab(sigp->signid+4, 4);
-    if (keyid == 0)
+    /* Just pretend we didn't see the keyid if we fail to lock */
+    if (pthread_mutex_lock(&keyid_lock))
 	return 0;
 
     if (keyids != NULL)
     for (i = 0; i < nkeyids; i++) {
 	if (keyid == keyids[i])
-	    return 1;
+	    seen = 1;
+	    goto exit;
     }
 
     if (nkeyids < nkeyids_max) {
@@ -176,207 +141,103 @@ static int stashKeyid(pgpDig dig)
     nextkeyid++;
     nextkeyid %= nkeyids_max;
 
-    return 0;
+exit:
+    pthread_mutex_unlock(&keyid_lock);
+    return seen;
 }
 
-/* Parse the parameters from the OpenPGP packets that will be needed. */
-static rpmRC parsePGP(rpmtd sigtd, const char *type, pgpDig dig)
+int parsePGPSig(rpmtd sigtd, const char *type,
+		 pgpDigParams *sig, char **msg)
 {
-    rpmRC rc = RPMRC_FAIL;
-    int debug = (_print_pkts & rpmIsDebug());
-    if ((pgpPrtPkts(sigtd->data, sigtd->count, dig, debug) == 0) &&
-	 (dig->signature.version == 3 || dig->signature.version == 4)) {
-	rc = RPMRC_OK;
-    } else {
-	rpmlog(RPMLOG_ERR,
-	    _("skipping %s with unverifiable V%u signature\n"), type,
-	    dig->signature.version);
-    }
+    int rc = pgpPrtParams(sigtd->data, sigtd->count, PGPTAG_SIGNATURE, sig);
+
+    if (rc != 0 && type != NULL)
+	rasprintf(msg, _("%s has unverifiable signature"), type);
     return rc;
 }
 
-static rpmRC headerVerify(rpmKeyring keyring, rpmVSFlags vsflags,
-			  const void * uh, size_t uc, char ** msg)
+/*
+ * Argument monster to verify header-only signature/digest if there is
+ * one, otherwisereturn RPMRC_NOTFOUND to signal for plain sanity check.
+ */
+static rpmRC headerSigVerify(rpmKeyring keyring, rpmVSFlags vsflags,
+			     int il, int dl, int ril, int rdl,
+			     entryInfo pe, unsigned char * dataStart,
+			     char **buf)
 {
-    pgpDig dig = NULL;
-    char *buf = NULL;
-    int32_t * ei = (int32_t *) uh;
-    int32_t il = ntohl(ei[0]);
-    int32_t dl = ntohl(ei[1]);
-    entryInfo pe = (entryInfo) &ei[2];
-    int32_t ildl[2];
-    int32_t pvlen = sizeof(ildl) + (il * sizeof(*pe)) + dl;
-    unsigned char * dataStart = (unsigned char *) (pe + il);
-    struct indexEntry_s entry;
-    struct entryInfo_s info;
-    unsigned const char * b;
     size_t siglen = 0;
-    size_t blen;
-    size_t nb;
-    int32_t ril = 0;
-    unsigned char * regionEnd = NULL;
-    rpmRC rc = RPMRC_FAIL;	/* assume failure */
-    int xx;
-    int i;
+    rpmRC rc = RPMRC_FAIL;
+    pgpDigParams sig = NULL;
     struct rpmtd_s sigtd;
-    DIGEST_CTX ctx = NULL;
+    struct entryInfo_s info, einfo;
+    unsigned int hashalgo = 0;
 
-    /* Is the blob the right size? */
-    if (uc > 0 && pvlen != uc) {
-	rasprintf(&buf, _("blob size(%d): BAD, 8 + 16 * il(%d) + dl(%d)\n"),
-		(int)uc, (int)il, (int)dl);
-	goto exit;
-    }
-
-    memset(&entry, 0, sizeof(entry));
+    rpmtdReset(&sigtd);
     memset(&info, 0, sizeof(info));
-
-    /* Check (and convert) the 1st tag element. */
-    xx = headerVerifyInfo(1, dl, pe, &entry.info, 0);
-    if (xx != -1) {
-	rasprintf(&buf, _("tag[%d]: BAD, tag %d type %d offset %d count %d\n"),
-		0, entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	goto exit;
-    }
-
-    /* Is there an immutable header region tag? */
-    if (!(entry.info.tag == RPMTAG_HEADERIMMUTABLE
-       && entry.info.type == RPM_BIN_TYPE
-       && entry.info.count == REGION_TAG_COUNT))
-    {
-	rc = RPMRC_NOTFOUND;
-	goto exit;
-    }
-
-    /* Is the offset within the data area? */
-    if (entry.info.offset >= dl) {
-	rasprintf(&buf, 
-		_("region offset: BAD, tag %d type %d offset %d count %d\n"),
-		entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	goto exit;
-    }
-
-    /* Is there an immutable header region tag trailer? */
-    regionEnd = dataStart + entry.info.offset;
-    (void) memcpy(&info, regionEnd, REGION_TAG_COUNT);
-    regionEnd += REGION_TAG_COUNT;
-
-    xx = headerVerifyInfo(1, dl, &info, &entry.info, 1);
-    if (xx != -1 ||
-	!(entry.info.tag == RPMTAG_HEADERIMMUTABLE
-       && entry.info.type == RPM_BIN_TYPE
-       && entry.info.count == REGION_TAG_COUNT))
-    {
-	rasprintf(&buf, 
-		_("region trailer: BAD, tag %d type %d offset %d count %d\n"),
-		entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	goto exit;
-    }
-    memset(&info, 0, sizeof(info));
-
-    /* Is the no. of tags in the region less than the total no. of tags? */
-    ril = entry.info.offset/sizeof(*pe);
-    if ((entry.info.offset % sizeof(*pe)) || ril > il) {
-	rasprintf(&buf, _("region size: BAD, ril(%d) > il(%d)\n"), ril, il);
-	goto exit;
-    }
+    memset(&einfo, 0, sizeof(einfo));
 
     /* Find a header-only digest/signature tag. */
-    for (i = ril; i < il; i++) {
-	xx = headerVerifyInfo(1, dl, pe+i, &entry.info, 0);
-	if (xx != -1) {
-	    rasprintf(&buf,
-		_("tag[%d]: BAD, tag %d type %d offset %d count %d\n"),
-		i, entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
+    for (int i = ril; i < il; i++) {
+	if (headerVerifyInfo(1, dl, pe+i, &einfo, 0) != -1) {
+	    rasprintf(buf,
+		_("tag[%d]: BAD, tag %d type %d offset %d count %d"),
+		i, einfo.tag, einfo.type,
+		einfo.offset, einfo.count);
 	    goto exit;
 	}
 
-	switch (entry.info.tag) {
-	case RPMTAG_SHA1HEADER:
+	switch (einfo.tag) {
+	case RPMTAG_SHA1HEADER: {
+	    size_t blen = 0;
+	    unsigned const char * b;
 	    if (vsflags & RPMVSF_NOSHA1HEADER)
 		break;
-	    blen = 0;
-	    for (b = dataStart + entry.info.offset; *b != '\0'; b++) {
+	    for (b = dataStart + einfo.offset; *b != '\0'; b++) {
 		if (strchr("0123456789abcdefABCDEF", *b) == NULL)
 		    break;
 		blen++;
 	    }
-	    if (entry.info.type != RPM_STRING_TYPE || *b != '\0' || blen != 40)
+	    if (einfo.type != RPM_STRING_TYPE || *b != '\0' || blen != 40)
 	    {
-		rasprintf(&buf, _("hdr SHA1: BAD, not hex\n"));
+		rasprintf(buf, _("hdr SHA1: BAD, not hex"));
 		goto exit;
 	    }
 	    if (info.tag == 0) {
-		info = entry.info;	/* structure assignment */
+		info = einfo;	/* structure assignment */
 		siglen = blen + 1;
 	    }
-	    break;
+	    } break;
 	case RPMTAG_RSAHEADER:
 	    if (vsflags & RPMVSF_NORSAHEADER)
 		break;
-	    if (entry.info.type != RPM_BIN_TYPE) {
-		rasprintf(&buf, _("hdr RSA: BAD, not binary\n"));
+	    if (einfo.type != RPM_BIN_TYPE) {
+		rasprintf(buf, _("hdr RSA: BAD, not binary"));
 		goto exit;
 	    }
-	    info = entry.info;	/* structure assignment */
+	    info = einfo;	/* structure assignment */
 	    siglen = info.count;
 	    break;
 	case RPMTAG_DSAHEADER:
 	    if (vsflags & RPMVSF_NODSAHEADER)
 		break;
-	    if (entry.info.type != RPM_BIN_TYPE) {
-		rasprintf(&buf, _("hdr DSA: BAD, not binary\n"));
+	    if (einfo.type != RPM_BIN_TYPE) {
+		rasprintf(buf, _("hdr DSA: BAD, not binary"));
 		goto exit;
 	    }
-	    info = entry.info;	/* structure assignment */
+	    info = einfo;	/* structure assignment */
 	    siglen = info.count;
 	    break;
 	default:
 	    break;
 	}
     }
-    rc = RPMRC_NOTFOUND;
 
-exit:
-    /* Return determined RPMRC_OK/RPMRC_FAIL conditions. */
-    if (rc != RPMRC_NOTFOUND) {
-	if (msg) 
-	    *msg = buf;
-	else
-	    free(buf);
-	return rc;
-    }
-
-    /* If no header-only digest/signature, then do simple sanity check. */
+    /* No header-only digest/signature found, get outta here */
     if (info.tag == 0) {
-verifyinfo_exit:
-	xx = headerVerifyInfo(ril-1, dl, pe+1, &entry.info, 0);
-	if (xx != -1) {
-	    rasprintf(&buf,
-		_("tag[%d]: BAD, tag %d type %d offset %d count %d\n"),
-		xx+1, entry.info.tag, entry.info.type,
-		entry.info.offset, entry.info.count);
-	    rc = RPMRC_FAIL;
-	} else {
-	    rasprintf(&buf, "Header sanity check: OK\n");
-	    rc = RPMRC_OK;
-	}
-	if (msg) 
-	    *msg = buf;
-	else
-	    free(buf);
-	return rc;
+	rc = RPMRC_NOTFOUND;
+	goto exit;
     }
 
-    /* Verify header-only digest/signature. */
-    dig = pgpNewDig();
-    if (dig == NULL)
-	goto verifyinfo_exit;
-
     sigtd.tag = info.tag;
     sigtd.type = info.type;
     sigtd.count = info.count;
@@ -386,51 +247,175 @@ verifyinfo_exit:
     switch (info.tag) {
     case RPMTAG_RSAHEADER:
     case RPMTAG_DSAHEADER:
-	if ((rc = parsePGP(&sigtd, "header", dig)) != RPMRC_OK) {
-	    pgpFreeDig(dig);
+	if (parsePGPSig(&sigtd, "header", &sig, buf))
 	    goto exit;
-	}
-	/* fallthrough */
-    case RPMTAG_SHA1HEADER: {
-	int hashalgo = (info.tag == RPMTAG_SHA1HEADER) ?
-			PGPHASHALGO_SHA1 : dig->signature.hash_algo;
-	ildl[0] = htonl(ril);
-	ildl[1] = (regionEnd - dataStart);
-	ildl[1] = htonl(ildl[1]);
-
-	ctx = rpmDigestInit(hashalgo, RPMDIGEST_NONE);
-
-	b = (unsigned char *) rpm_header_magic;
-	nb = sizeof(rpm_header_magic);
-        (void) rpmDigestUpdate(ctx, b, nb);
-
-	b = (unsigned char *) ildl;
-	nb = sizeof(ildl);
-        (void) rpmDigestUpdate(ctx, b, nb);
-
-	b = (unsigned char *) pe;
-	nb = (htonl(ildl[0]) * sizeof(*pe));
-        (void) rpmDigestUpdate(ctx, b, nb);
-
-	b = (unsigned char *) dataStart;
-	nb = htonl(ildl[1]);
-        (void) rpmDigestUpdate(ctx, b, nb);
-	} break;
+	hashalgo = pgpDigParamsAlgo(sig, PGPVAL_HASHALGO);
+	break;
+    case RPMTAG_SHA1HEADER:
+	hashalgo = PGPHASHALGO_SHA1;
+	break;
     default:
-	sigtd.data = _free(sigtd.data); /* Hmm...? */
 	break;
     }
 
-    rc = rpmVerifySignature(keyring, &sigtd, dig, ctx, &buf);
+    if (hashalgo) {
+	DIGEST_CTX ctx = rpmDigestInit(hashalgo, RPMDIGEST_NONE);
+	int32_t ildl[2] = { htonl(ril), htonl(rdl) };
+
+	rpmDigestUpdate(ctx, rpm_header_magic, sizeof(rpm_header_magic));
+	rpmDigestUpdate(ctx, ildl, sizeof(ildl));
+	rpmDigestUpdate(ctx, pe, (ril * sizeof(*pe)));
+	rpmDigestUpdate(ctx, dataStart, rdl);
+
+	rc = rpmVerifySignature(keyring, &sigtd, sig, ctx, buf);
+
+    	rpmDigestFinal(ctx, NULL, NULL, 0);
+    }
+
+exit:
+    rpmtdFreeData(&sigtd);
+    pgpDigParamsFree(sig);
+
+    return rc;
+}
+
+rpmRC headerVerifyRegion(rpmTagVal regionTag,
+			struct indexEntry_s *entry, int il, int dl,
+			entryInfo pe, unsigned char *dataStart,
+			int *rilp, int *rdlp, char **buf)
+{
+    rpmRC rc = RPMRC_FAIL;
+    struct entryInfo_s info;
+    unsigned char * regionEnd = NULL;
+    int32_t ril = 0;
+    int32_t rdl = 0;
+
+    memset(entry, 0, sizeof(*entry));
+
+    /* Check (and convert) the 1st tag element. */
+    if (headerVerifyInfo(1, dl, pe, &entry->info, 0) != -1) {
+	rasprintf(buf, _("tag[%d]: BAD, tag %d type %d offset %d count %d"),
+		0, entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is there an immutable header region tag? */
+    if (!(entry->info.tag == regionTag)) {
+	rc = RPMRC_NOTFOUND;
+	goto exit;
+    }
+
+    /* Is the region tag sane? */
+    if (!(entry->info.type == REGION_TAG_TYPE &&
+	  entry->info.count == REGION_TAG_COUNT)) {
+	rasprintf(buf,
+		_("region tag: BAD, tag %d type %d offset %d count %d"),
+		entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is the trailer within the data area? */
+    if (entry->info.offset + REGION_TAG_COUNT > dl) {
+	rasprintf(buf, 
+		_("region offset: BAD, tag %d type %d offset %d count %d"),
+		entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is there an immutable header region tag trailer? */
+    memset(&info, 0, sizeof(info));
+    regionEnd = dataStart + entry->info.offset;
+    (void) memcpy(&info, regionEnd, REGION_TAG_COUNT);
+    regionEnd += REGION_TAG_COUNT;
+    rdl = regionEnd - dataStart;
+
+    if (headerVerifyInfo(1, il * sizeof(*pe), &info, &entry->info, 1) != -1 ||
+	!(entry->info.tag == regionTag
+       && entry->info.type == REGION_TAG_TYPE
+       && entry->info.count == REGION_TAG_COUNT))
+    {
+	rasprintf(buf, 
+		_("region trailer: BAD, tag %d type %d offset %d count %d"),
+		entry->info.tag, entry->info.type,
+		entry->info.offset, entry->info.count);
+	goto exit;
+    }
+
+    /* Is the no. of tags in the region less than the total no. of tags? */
+    ril = entry->info.offset/sizeof(*pe);
+    if ((entry->info.offset % sizeof(*pe)) || ril > il) {
+	rasprintf(buf, _("region size: BAD, ril(%d) > il(%d)"), ril, il);
+	goto exit;
+    }
+
+    rc = RPMRC_OK;
+    if (rilp)
+	*rilp = ril;
+    if (rdlp)
+	*rdlp = rdl;
+
+exit:
+    return rc;
+}
+
+static rpmRC headerVerify(rpmKeyring keyring, rpmVSFlags vsflags,
+			  const void * uh, size_t uc, char ** msg)
+{
+    char *buf = NULL;
+    int32_t * ei = (int32_t *) uh;
+    int32_t il = ntohl(ei[0]);
+    int32_t dl = ntohl(ei[1]);
+    entryInfo pe = (entryInfo) &ei[2];
+    int32_t pvlen = sizeof(il) + sizeof(dl) + (il * sizeof(*pe)) + dl;
+    unsigned char * dataStart = (unsigned char *) (pe + il);
+    struct indexEntry_s entry;
+    int32_t ril = 0;
+    int32_t rdl = 0;
+    rpmRC rc = RPMRC_FAIL;	/* assume failure */
+
+    /* Is the blob the right size? */
+    if (uc > 0 && pvlen != uc) {
+	rasprintf(&buf, _("blob size(%d): BAD, 8 + 16 * il(%d) + dl(%d)"),
+		(int)uc, (int)il, (int)dl);
+	goto exit;
+    }
+
+    /* Verify header immutable region if there is one */
+    rc = headerVerifyRegion(RPMTAG_HEADERIMMUTABLE,
+			    &entry, il, dl, pe, dataStart,
+			    &ril, &rdl, &buf);
+
+    /* Verify header-only digest/signature if there is one we can use. */
+    if (rc == RPMRC_OK) {
+	rc = headerSigVerify(keyring, vsflags,
+			     il, dl, ril, rdl,
+			     pe, dataStart, &buf);
+    }
+
+exit:
+    /* If no header-only digest/signature, then do simple sanity check. */
+    if (rc == RPMRC_NOTFOUND) {
+	int xx = headerVerifyInfo(ril-1, dl, pe+1, &entry.info, 0);
+	if (xx != -1) {
+	    rasprintf(&buf,
+		_("tag[%d]: BAD, tag %d type %d offset %d count %d"),
+		xx+1, entry.info.tag, entry.info.type,
+		entry.info.offset, entry.info.count);
+	    rc = RPMRC_FAIL;
+	} else {
+	    rasprintf(&buf, "Header sanity check: OK");
+	    rc = RPMRC_OK;
+	}
+    }
 
     if (msg) 
 	*msg = buf;
     else
 	free(buf);
 
-    rpmtdFreeData(&sigtd);
-    pgpFreeDig(dig);
-    rpmDigestFinal(ctx, NULL, NULL, 0);
     return rc;
 }
 
@@ -468,24 +453,24 @@ static rpmRC rpmpkgReadHeader(rpmKeyring keyring, rpmVSFlags vsflags,
 	*msg = NULL;
 
     memset(block, 0, sizeof(block));
-    if ((xx = timedRead(fd, (char *)block, sizeof(block))) != sizeof(block)) {
+    if ((xx = Freadall(fd, block, sizeof(block))) != sizeof(block)) {
 	rasprintf(&buf, 
-		_("hdr size(%d): BAD, read returned %d\n"), (int)sizeof(block), xx);
+		_("hdr size(%d): BAD, read returned %d"), (int)sizeof(block), xx);
 	goto exit;
     }
     if (memcmp(block, rpm_header_magic, sizeof(rpm_header_magic))) {
-	rasprintf(&buf, _("hdr magic: BAD\n"));
+	rasprintf(&buf, _("hdr magic: BAD"));
 	goto exit;
     }
     il = ntohl(block[2]);
     if (hdrchkTags(il)) {
-	rasprintf(&buf, _("hdr tags: BAD, no. of tags(%d) out of range\n"), il);
+	rasprintf(&buf, _("hdr tags: BAD, no. of tags(%d) out of range"), il);
 	goto exit;
     }
     dl = ntohl(block[3]);
     if (hdrchkData(dl)) {
 	rasprintf(&buf,
-		  _("hdr data: BAD, no. of bytes(%d) out of range\n"), dl);
+		  _("hdr data: BAD, no. of bytes(%d) out of range"), dl);
 	goto exit;
     }
 
@@ -494,20 +479,21 @@ static rpmRC rpmpkgReadHeader(rpmKeyring keyring, rpmVSFlags vsflags,
     ei = xmalloc(uc);
     ei[0] = block[2];
     ei[1] = block[3];
-    if ((xx = timedRead(fd, (char *)&ei[2], nb)) != nb) {
-	rasprintf(&buf, _("hdr blob(%zd): BAD, read returned %d\n"), nb, xx);
+    if ((xx = Freadall(fd, (char *)&ei[2], nb)) != nb) {
+	rasprintf(&buf, _("hdr blob(%zd): BAD, read returned %d"), nb, xx);
 	goto exit;
     }
 
     /* Sanity check header tags */
-    rc = headerVerify(keyring, vsflags, ei, uc, msg);
+    rc = headerVerify(keyring, vsflags, ei, uc, &buf);
     if (rc != RPMRC_OK)
 	goto exit;
 
     /* OK, blob looks sane, load the header. */
-    h = headerLoad(ei);
+    h = headerImport(ei, uc, 0);
     if (h == NULL) {
-	rasprintf(&buf, _("hdr load: BAD\n"));
+	free(buf);
+	rasprintf(&buf, _("hdr load: BAD"));
 	rc = RPMRC_FAIL;
         goto exit;
     }
@@ -516,8 +502,8 @@ static rpmRC rpmpkgReadHeader(rpmKeyring keyring, rpmVSFlags vsflags,
 exit:
     if (hdrp && h && rc == RPMRC_OK)
 	*hdrp = headerLink(h);
-    ei = _free(ei);
-    h = headerFree(h);
+    free(ei);
+    headerFree(h);
 
     if (msg != NULL && *msg == NULL && buf != NULL) {
 	*msg = buf;
@@ -541,17 +527,16 @@ rpmRC rpmReadHeader(rpmts ts, FD_t fd, Header *hdrp, char ** msg)
 }
 
 static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags, 
-			FD_t fd, const char * fn, Header * hdrp)
+			FD_t fd,
+			Header * hdrp, unsigned int *keyidp, char **msg)
 {
-    pgpDig dig = NULL;
+    pgpDigParams sig = NULL;
     char buf[8*BUFSIZ];
     ssize_t count;
-    rpmlead l = NULL;
     Header sigh = NULL;
     rpmTagVal sigtag;
     struct rpmtd_s sigtd;
     Header h = NULL;
-    char * msg;
     rpmRC rc = RPMRC_FAIL;	/* assume failure */
     int leadtype = -1;
     headerGetFlags hgeflags = HEADERGET_DEFAULT;
@@ -560,39 +545,21 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
     if (hdrp) *hdrp = NULL;
 
     rpmtdReset(&sigtd);
-    l = rpmLeadNew();
 
-    if ((rc = rpmLeadRead(fd, l)) == RPMRC_OK) {
-	const char * err = NULL;
-	if ((rc = rpmLeadCheck(l, &err)) == RPMRC_FAIL) {
-	    rpmlog(RPMLOG_ERR, "%s: %s\n", fn, err);
+    if ((rc = rpmLeadRead(fd, NULL, &leadtype, msg)) != RPMRC_OK) {
+	/* Avoid message spew on manifests */
+	if (rc == RPMRC_NOTFOUND) {
+	    *msg = _free(*msg);
 	}
-	leadtype = rpmLeadType(l);
-    }
-    l = rpmLeadFree(l);
-
-    if (rc != RPMRC_OK)
 	goto exit;
+    }
 
     /* Read the signature header. */
-    msg = NULL;
-    rc = rpmReadSignature(fd, &sigh, RPMSIGTYPE_HEADERSIG, &msg);
-    switch (rc) {
-    default:
-	rpmlog(RPMLOG_ERR, _("%s: rpmReadSignature failed: %s"), fn,
-		(msg && *msg ? msg : "\n"));
-	msg = _free(msg);
+    rc = rpmReadSignature(fd, &sigh, RPMSIGTYPE_HEADERSIG, msg);
+
+    if (rc != RPMRC_OK) {
 	goto exit;
-	break;
-    case RPMRC_OK:
-	if (sigh == NULL) {
-	    rpmlog(RPMLOG_ERR, _("%s: No signature available\n"), fn);
-	    rc = RPMRC_FAIL;
-	    goto exit;
-	}
-	break;
     }
-    msg = _free(msg);
 
 #define	_chk(_mask, _tag) \
 	(sigtag == 0 && !(vsflags & (_mask)) && headerIsEntry(sigh, (_tag)))
@@ -623,17 +590,12 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 
     /* Read the metadata, computing digest(s) on the fly. */
     h = NULL;
-    msg = NULL;
 
-    rc = rpmpkgReadHeader(keyring, vsflags, fd, &h, &msg);
+    rc = rpmpkgReadHeader(keyring, vsflags, fd, &h, msg);
 
     if (rc != RPMRC_OK || h == NULL) {
-	rpmlog(RPMLOG_ERR, _("%s: headerRead failed: %s"), fn,
-		(msg && *msg ? msg : "\n"));
-	msg = _free(msg);
 	goto exit;
     }
-    msg = _free(msg);
 
     /* Any digests or signatures to check? */
     if (sigtag == 0) {
@@ -641,11 +603,8 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 	goto exit;
     }
 
-    dig = pgpNewDig();
-    if (dig == NULL) {
-	rc = RPMRC_FAIL;
-	goto exit;
-    }
+    /* Free up any previous "ok" message before signature/digest check */
+    *msg = _free(*msg);
 
     /* Retrieve the tag parameters from the signature header. */
     if (!headerGet(sigh, sigtag, &sigtd, hgeflags)) {
@@ -656,14 +615,16 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
     switch (sigtag) {
     case RPMSIGTAG_RSA:
     case RPMSIGTAG_DSA:
-	if ((rc = parsePGP(&sigtd, "package", dig)) != RPMRC_OK) {
+	if (parsePGPSig(&sigtd, "package", &sig, msg)) {
+	    rc = RPMRC_FAIL;
 	    goto exit;
 	}
 	/* fallthrough */
     case RPMSIGTAG_SHA1:
     {	struct rpmtd_s utd;
-	int hashalgo = (sigtag == RPMSIGTAG_SHA1) ?
-			PGPHASHALGO_SHA1 : dig->signature.hash_algo;
+	unsigned int hashalgo = (sigtag == RPMSIGTAG_SHA1) ?
+				PGPHASHALGO_SHA1 :
+				pgpDigParamsAlgo(sig, PGPVAL_HASHALGO);
 
 	if (!headerGet(h, RPMTAG_HEADERIMMUTABLE, &utd, hgeflags))
 	    break;
@@ -675,7 +636,8 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
     case RPMSIGTAG_GPG:
     case RPMSIGTAG_PGP5:	/* XXX legacy */
     case RPMSIGTAG_PGP:
-	if ((rc = parsePGP(&sigtd, "package", dig)) != RPMRC_OK) {
+	if (parsePGPSig(&sigtd, "package", &sig, msg)) {
+	    rc = RPMRC_FAIL;
 	    goto exit;
 	}
 	/* fallthrough */
@@ -683,41 +645,21 @@ static rpmRC rpmpkgRead(rpmKeyring keyring, rpmVSFlags vsflags,
 	/* Legacy signatures need the compressed payload in the digest too. */
 	while ((count = Fread(buf, sizeof(buf[0]), sizeof(buf), fd)) > 0) {}
 	if (count < 0) {
-	    rpmlog(RPMLOG_ERR, _("%s: Fread failed: %s\n"),
-					fn, Fstrerror(fd));
+	    rasprintf(msg, _("Fread failed: %s"), Fstrerror(fd));
 	    rc = RPMRC_FAIL;
 	    goto exit;
 	}
 
-	ctx = rpmDigestBundleDupCtx(fdGetBundle(fd), (sigtag == RPMSIGTAG_MD5) ?
-				    PGPHASHALGO_MD5 : dig->signature.hash_algo);
+	ctx = rpmDigestBundleDupCtx(fdGetBundle(fd),(sigtag == RPMSIGTAG_MD5) ?
+				    PGPHASHALGO_MD5 :
+				    pgpDigParamsAlgo(sig, PGPVAL_HASHALGO));
 	break;
     default:
 	break;
     }
 
     /** @todo Implement disable/enable/warn/error/anal policy. */
-    rc = rpmVerifySignature(keyring, &sigtd, dig, ctx, &msg);
-	
-    switch (rc) {
-    case RPMRC_OK:		/* Signature is OK. */
-	rpmlog(RPMLOG_DEBUG, "%s: %s", fn, msg);
-	break;
-    case RPMRC_NOTTRUSTED:	/* Signature is OK, but key is not trusted. */
-    case RPMRC_NOKEY:		/* Public key is unavailable. */
-	/* XXX Print NOKEY/NOTTRUSTED warning only once. */
-    {	int lvl = (stashKeyid(dig) ? RPMLOG_DEBUG : RPMLOG_WARNING);
-	rpmlog(lvl, "%s: %s", fn, msg);
-    }	break;
-    case RPMRC_NOTFOUND:	/* Signature is unknown type. */
-	rpmlog(RPMLOG_WARNING, "%s: %s", fn, msg);
-	break;
-    default:
-    case RPMRC_FAIL:		/* Signature does not verify. */
-	rpmlog(RPMLOG_ERR, "%s: %s", fn, msg);
-	break;
-    }
-    free(msg);
+    rc = rpmVerifySignature(keyring, &sigtd, sig, ctx, msg);
 
 exit:
     if (rc != RPMRC_FAIL && h != NULL && hdrp != NULL) {
@@ -736,23 +678,29 @@ exit:
 	    headerPutString(h, RPMTAG_SOURCERPM, "(none)");
 	}
 	/* 
-         * Convert legacy headers on the fly. Not having "new" style compressed
-         * filenames is close enough estimate for legacy indication... 
+         * Convert legacy headers on the fly. Not having immutable region
+         * equals a truly ancient package, do full retrofit. OTOH newer
+         * packages might have been built with --nodirtokens, test and handle
+         * the non-compressed filelist case separately.
          */
-	if (!headerIsEntry(h, RPMTAG_DIRNAMES)) {
+	if (!headerIsEntry(h, RPMTAG_HEADERIMMUTABLE))
 	    headerConvert(h, HEADERCONV_RETROFIT_V3);
-	}
+	else if (headerIsEntry(h, RPMTAG_OLDFILENAMES))
+	    headerConvert(h, HEADERCONV_COMPRESSFILELIST);
 	
 	/* Append (and remap) signature tags to the metadata. */
 	headerMergeLegacySigs(h, sigh);
 
 	/* Bump reference count for return. */
 	*hdrp = headerLink(h);
+
+	if (keyidp)
+	    *keyidp = getKeyid(sig);
     }
     rpmtdFreeData(&sigtd);
     rpmDigestFinal(ctx, NULL, NULL, 0);
     h = headerFree(h);
-    pgpFreeDig(dig);
+    pgpDigParamsFree(sig);
     sigh = rpmFreeSignature(sigh);
     return rc;
 }
@@ -761,47 +709,40 @@ rpmRC rpmReadPackageFile(rpmts ts, FD_t fd, const char * fn, Header * hdrp)
 {
     rpmRC rc;
     rpmVSFlags vsflags = rpmtsVSFlags(ts);
-    rpmKeyring keyring = 0;
+    rpmKeyring keyring = rpmtsGetKeyring(ts, 1);
+    unsigned int keyid = 0;
+    char *msg = NULL;
 
-    if ((vsflags & _RPMVSF_NOSIGNATURES) != _RPMVSF_NOSIGNATURES)
-	keyring = rpmtsGetKeyring(ts, 1);
+    if (fn == NULL)
+	fn = Fdescr(fd);
 
-    rc = rpmpkgRead(keyring, vsflags, fd, fn, hdrp);
+    rc = rpmpkgRead(keyring, vsflags, fd, hdrp, &keyid, &msg);
 
-    if (keyring)
-	rpmKeyringFree(keyring);
-    return rc;
-}
-
-/**
- * Check for supported payload format in header.
- * @param h		header to check
- * @return		RPMRC_OK if supported, RPMRC_FAIL otherwise
- */
-rpmRC headerCheckPayloadFormat(Header h) {
-    rpmRC rc = RPMRC_OK;
-    const char *payloadfmt = headerGetString(h, RPMTAG_PAYLOADFORMAT);
-    /* 
-     * XXX Ugh, rpm 3.x packages don't have payload format tag. Instead
-     * of blinly allowing, should check somehow (HDRID existence or... ?)
-     */
-    if (!payloadfmt) return rc;
-
-    if (!rstreq(payloadfmt, "cpio")) {
-        char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
-        if (payloadfmt && rstreq(payloadfmt, "drpm")) {
-            rpmlog(RPMLOG_ERR,
-                     _("%s is a Delta RPM and cannot be directly installed\n"),
-                     nevra);
-        } else {
-            rpmlog(RPMLOG_ERR, 
-                     _("Unsupported payload (%s) in package %s\n"),
-                     payloadfmt ? payloadfmt : "none", nevra);
-        } 
-        nevra = _free(nevra);
-	rc = RPMRC_FAIL;
+    switch (rc) {
+    case RPMRC_OK:		/* Signature is OK. */
+	rpmlog(RPMLOG_DEBUG, "%s: %s\n", fn, msg);
+	break;
+    case RPMRC_NOTTRUSTED:	/* Signature is OK, but key is not trusted. */
+    case RPMRC_NOKEY:		/* Public key is unavailable. */
+	/* XXX Print NOKEY/NOTTRUSTED warning only once. */
+    {	int lvl = (stashKeyid(keyid) ? RPMLOG_DEBUG : RPMLOG_WARNING);
+	rpmlog(lvl, "%s: %s\n", fn, msg);
+    }	break;
+    case RPMRC_NOTFOUND:	/* Signature is unknown type or manifest. */
+	/* msg == NULL is probably a manifest */
+	if (msg)
+	    rpmlog(RPMLOG_WARNING, "%s: %s\n", fn, msg);
+	break;
+    default:
+    case RPMRC_FAIL:		/* Signature does not verify. */
+	rpmlog(RPMLOG_ERR, "%s: %s\n", fn, msg);
+	break;
     }
+    rpmKeyringFree(keyring);
+    free(msg);
+
     return rc;
 }
 
 
+
