@@ -6,10 +6,6 @@
 #include "system.h"
 const char *__progname;
 
-#if HAVE_MCHECK_H
-#include <mcheck.h>
-#endif
-
 #include <rpm/rpmcli.h>
 #include <rpm/rpmlib.h>		/* rpmEVR, rpmReadConfigFiles etc */
 #include <rpm/rpmlog.h>
@@ -23,6 +19,8 @@ const char *__progname;
 #define POPT_QUERYTAGS		-997
 #define POPT_PREDEFINE		-996
 #define POPT_DBPATH		-995
+#define POPT_UNDEFINE		-994
+#define POPT_PIPE		-993
 
 static int _debug = 0;
 
@@ -108,14 +106,20 @@ static void rpmcliAllArgCallback( poptContext con,
 	rpmcliConfigured();
 	(void) rpmDefineMacro(NULL, t, RMIL_CMDLINE);
 	(void) rpmDefineMacro(rpmCLIMacroContext, t, RMIL_CMDLINE);
-	s = _free(s);
+	free(s);
 	break;
     }
+    case POPT_UNDEFINE:
+	rpmcliConfigured();
+	if (*arg == '%')
+	    arg++;
+	delMacro(NULL, arg);
+	break;
     case 'E':
 	rpmcliConfigured();
 	{   char *val = rpmExpand(arg, NULL);
 	    fprintf(stdout, "%s\n", val);
-	    val = _free(val);
+	    free(val);
 	}
 	break;
     case POPT_DBPATH:
@@ -135,6 +139,16 @@ static void rpmcliAllArgCallback( poptContext con,
 	rpmDisplayQueryTags(stdout);
 	exit(EXIT_SUCCESS);
 	break;
+    case POPT_PIPE:
+	if (rpmcliPipeOutput) {
+	    fprintf(stderr,
+		    _("%s: error: more than one --pipe specified "
+		      "(incompatible popt aliases?)\n"), __progname);
+	    exit(EXIT_FAILURE);
+	}
+	rpmcliPipeOutput = xstrdup(arg);
+	break;
+	
     case RPMCLI_POPT_NODIGEST:
 	rpmcliQueryFlags |= VERIFY_DIGEST;
 	break;
@@ -163,6 +177,9 @@ struct poptOption rpmcliAllPoptTable[] = {
  { "define", 'D', POPT_ARG_STRING, 0, 'D',
 	N_("define MACRO with value EXPR"),
 	N_("'MACRO EXPR'") },
+ { "undefine", '\0', POPT_ARG_STRING, 0, POPT_UNDEFINE,
+	N_("undefine MACRO"),
+	N_("MACRO") },
  { "eval", 'E', POPT_ARG_STRING, 0, 'E',
 	N_("print macro expansion of EXPR"),
 	N_("'EXPR'") },
@@ -170,6 +187,11 @@ struct poptOption rpmcliAllPoptTable[] = {
 	N_("read <FILE:...> instead of default file(s)"),
 	N_("<FILE:...>") },
 
+ /* XXX this is a bit out of place here but kinda unavoidable... */
+ { "noplugins", '\0', POPT_BIT_SET,
+	&rpmIArgs.transFlags, RPMTRANS_FLAG_NOPLUGINS,
+	N_("don't enable any plugins"), NULL },
+
  { "nodigest", '\0', 0, 0, RPMCLI_POPT_NODIGEST,
         N_("don't verify package digest(s)"), NULL },
  { "nohdrchk", '\0', POPT_ARGFLAG_DOC_HIDDEN, 0, RPMCLI_POPT_NOHDRCHK,
@@ -177,7 +199,7 @@ struct poptOption rpmcliAllPoptTable[] = {
  { "nosignature", '\0', 0, 0, RPMCLI_POPT_NOSIGNATURE,
         N_("don't verify package signature(s)"), NULL },
 
- { "pipe", '\0', POPT_ARG_STRING|POPT_ARGFLAG_DOC_HIDDEN, &rpmcliPipeOutput, 0,
+ { "pipe", '\0', POPT_ARG_STRING|POPT_ARGFLAG_DOC_HIDDEN, 0, POPT_PIPE,
 	N_("send stdout to CMD"),
 	N_("CMD") },
  { "rcfile", '\0', POPT_ARG_STRING, &rpmcliRcfile, 0,
@@ -228,24 +250,16 @@ rpmcliFini(poptContext optCon)
     rpmlogClose();
     rpmcliInitialized = -1;
 
-#if HAVE_MCHECK_H && HAVE_MTRACE
-    muntrace();   /* Trace malloc only if MALLOC_TRACE=mtrace-output-file. */
-#endif
-
     return NULL;
 }
 
 poptContext
 rpmcliInit(int argc, char *const argv[], struct poptOption * optionsTable)
 {
-    const char * optArg;
     poptContext optCon;
     int rc;
     const char *ctx, *execPath;
 
-#if HAVE_MCHECK_H && HAVE_MTRACE
-    mtrace();   /* Trace malloc only if MALLOC_TRACE=mtrace-output-file. */
-#endif
     setprogname(argv[0]);       /* Retrofit glibc __progname */
 
     /* XXX glibc churn sanity */
@@ -286,15 +300,9 @@ rpmcliInit(int argc, char *const argv[], struct poptOption * optionsTable)
 
     /* Process all options, whine if unknown. */
     while ((rc = poptGetNextOpt(optCon)) > 0) {
-	optArg = poptGetOptArg(optCon);
-	switch (rc) {
-	default:
-	    fprintf(stderr, _("%s: option table misconfigured (%d)\n"),
+	fprintf(stderr, _("%s: option table misconfigured (%d)\n"),
 		__progname, rc);
-	    exit(EXIT_FAILURE);
-
-	    break;
-        }
+	exit(EXIT_FAILURE);
     }
 
     if (rc < -1) {
