@@ -35,23 +35,6 @@ extern const char * const rpmEVR;
 
 extern const int rpmFLAGS;
 
-/** \ingroup header
- * Translate and merge legacy signature tags into header.
- * @todo Remove headerSort() through headerInitIterator() modifies sig.
- * @param h		header
- * @param sigh		signature header
- */
-void headerMergeLegacySigs(Header h, const Header sigh);
-
-/** \ingroup header
- * Regenerate signature header.
- * @todo Remove headerSort() through headerInitIterator() modifies h.
- * @param h		header
- * @param noArchiveSize	don't copy archive size tag (pre rpm-4.1)
- * @return		regenerated signature header
- */
-Header headerRegenSigHeader(const Header h, int noArchiveSize);
-
 /* ==================================================================== */
 /** \name RPMRC */
 
@@ -85,6 +68,13 @@ int rpmReadConfigFiles(const char * file,
 void rpmGetArchInfo( const char ** name,
 		int * num);
 
+/** \ingroup rpmrc
+ * Return color for an arch
+ * @param arch		name of an architecture
+ * @return color        color of arch, -1 if the arch couldn't be determined
+ */
+int rpmGetArchColor(const char *arch);
+
 /** \ingroup rpmrc
  * Return current os name and/or number.
  * @todo Generalize to extract os component from target_platform macro.
@@ -116,15 +106,6 @@ int rpmMachineScore(int type, const char * name);
  */
 int rpmShowRC(FILE * fp);
 
-/** \ingroup rpmrc
- * @deprecated Use addMacro to set _target_* macros.
- * @todo Eliminate from API.
- # @note Only used by build code.
- * @param archTable
- * @param osTable
- */
-void rpmSetTables(int archTable, int osTable);
-
 /** \ingroup rpmrc
  * Destroy rpmrc arch/os compatibility tables.
  * @todo Eliminate from API.
@@ -139,13 +120,6 @@ void rpmFreeRpmrc(void);
  */
 int rpmVersionCompare(Header first, Header second);
 
-/** \ingroup header
- * Check for supported payload format in header.
- * @param h		header to check
- * @return		RPMRC_OK if supported, RPMRC_FAIL otherwise
- */
-rpmRC headerCheckPayloadFormat(Header h);
-
 /**  \ingroup header
  * Check header consistency, performing headerGetEntry() the hard way.
  *  
