@@ -1,14 +1,17 @@
 #include "system.h"
 
 #include <stdarg.h>
-#if defined(__linux__) && defined(__powerpc__)
-#include <setjmp.h>
+#include <pthread.h>
+
+#if defined(__linux__)
+#include <elf.h>
+#include <link.h>
 #endif
 
+
 #if HAVE_SYS_UTSNAME_H
 #include <sys/utsname.h>
 #endif
-#include <netdb.h>
 #include <ctype.h>	/* XXX for /etc/rpm/platform contents */
 
 #if HAVE_SYS_SYSTEMCFG_H
@@ -17,32 +20,32 @@
 #define __power_pc() 0
 #endif
 
+#ifdef HAVE_SYS_AUXV_H
+#include <sys/auxv.h>
+#endif
+
 #include <rpm/rpmlib.h>			/* RPM_MACTABLE*, Rc-prototypes */
 #include <rpm/rpmmacro.h>
 #include <rpm/rpmfileutil.h>
 #include <rpm/rpmstring.h>
 #include <rpm/rpmlog.h>
+#include <rpm/argv.h>
 
 #include "rpmio/rpmlua.h"
 #include "rpmio/rpmio_internal.h"	/* XXX for rpmioSlurp */
 #include "lib/misc.h"
 #include "lib/rpmliblua.h"
+#include "lib/rpmug.h"
 
 #include "debug.h"
 
 static const char * defrcfiles = NULL;
 const char * macrofiles = NULL;
 
-static const char * const platform = SYSCONFDIR "/rpm/platform";
-static char ** platpat = NULL;
-static int nplatpat = 0;
-
-typedef char * cptr_t;
-
 typedef struct machCacheEntry_s {
     char * name;
     int count;
-    cptr_t * equivs;
+    char ** equivs;
     int visited;
 } * machCacheEntry;
 
@@ -72,12 +75,15 @@ struct rpmOption {
     char * name;
     int var;
     int archSpecific;
-int required;
     int macroize;
     int localize;
-struct rpmOptionValue * value;
 };
 
+static struct rpmat_s {
+    const char *platform;
+    uint64_t hwcap;
+} rpmat;
+
 typedef struct defaultEntry_s {
     char * name;
     char * defName;
@@ -105,28 +111,23 @@ typedef struct tableType_s {
     int canonsLength;
 } * tableType;
 
-static struct tableType_s tables[RPM_MACHTABLE_COUNT] = {
-    { "arch", 1, 0 },
-    { "os", 1, 0 },
-    { "buildarch", 0, 1 },
-    { "buildos", 0, 1 }
-};
-
 /* XXX get rid of this stuff... */
 /* Stuff for maintaining "variables" like SOURCEDIR, BUILDDIR, etc */
 #define RPMVAR_OPTFLAGS                 3
+#define RPMVAR_ARCHCOLOR                42
 #define RPMVAR_INCLUDE                  43
 #define RPMVAR_MACROFILES               49
 
 #define RPMVAR_NUM                      55      /* number of RPMVAR entries */
 
 /* this *must* be kept in alphabetical order */
-/* The order of the flags is archSpecific, required, macroize, localize */
+/* The order of the flags is archSpecific, macroize, localize */
 
 static const struct rpmOption optionTable[] = {
-    { "include",		RPMVAR_INCLUDE,			0, 1,	0, 2 },
-    { "macrofiles",		RPMVAR_MACROFILES,		0, 0,	0, 1 },
-    { "optflags",		RPMVAR_OPTFLAGS,		1, 0,	1, 0 },
+    { "archcolor",		RPMVAR_ARCHCOLOR,               1, 0, 0 },
+    { "include",		RPMVAR_INCLUDE,			0, 0, 2 },
+    { "macrofiles",		RPMVAR_MACROFILES,		0, 0, 1 },
+    { "optflags",		RPMVAR_OPTFLAGS,		1, 1, 0 },
 };
 
 static const size_t optionTableSize = sizeof(optionTable) / sizeof(*optionTable);
@@ -134,23 +135,58 @@ static const size_t optionTableSize = sizeof(optionTable) / sizeof(*optionTable)
 #define OS	0
 #define ARCH	1
 
-static cptr_t current[2];
-
-static int currTables[2] = { RPM_MACHTABLE_INSTOS, RPM_MACHTABLE_INSTARCH };
-
-static struct rpmvarValue values[RPMVAR_NUM];
-
 static int defaultsInitialized = 0;
 
+typedef struct rpmrcCtx_s * rpmrcCtx;
+struct rpmrcCtx_s {
+    ARGV_t platpat;
+    char *current[2];
+    int currTables[2];
+    struct rpmvarValue values[RPMVAR_NUM];
+    struct tableType_s tables[RPM_MACHTABLE_COUNT];
+    pthread_rwlock_t lock;
+};
+
 /* prototypes */
-static rpmRC doReadRC(const char * urlfn);
+static rpmRC doReadRC(rpmrcCtx ctx, const char * urlfn);
 
-static void rpmSetVarArch(int var, const char * val,
-		const char * arch);
+static void rpmSetVarArch(rpmrcCtx ctx,
+			  int var, const char * val, const char * arch);
 
-static void rebuildCompatTables(int type, const char * name);
+static void rebuildCompatTables(rpmrcCtx ctx, int type, const char * name);
 
-static void rpmRebuildTargetVars(const char **target, const char ** canontarget);
+static void rpmRebuildTargetVars(rpmrcCtx ctx, const char **target, const char ** canontarget);
+
+/* Force context (lock) acquisition through a function */
+static rpmrcCtx rpmrcCtxAcquire(int write)
+{
+    static struct rpmrcCtx_s _globalCtx = {
+	.lock = PTHREAD_RWLOCK_INITIALIZER,
+	.currTables = { RPM_MACHTABLE_INSTOS, RPM_MACHTABLE_INSTARCH },
+	.tables = {
+	    { "arch", 1, 0 },
+	    { "os", 1, 0 },
+	    { "buildarch", 0, 1 },
+	    { "buildos", 0, 1 }
+	},
+    };
+    rpmrcCtx ctx = &_globalCtx;
+
+    /* XXX: errors should be handled */
+    if (write)
+	pthread_rwlock_wrlock(&ctx->lock);
+    else
+	pthread_rwlock_rdlock(&ctx->lock);
+
+    return ctx;
+}
+
+/* Release context (lock) */
+static rpmrcCtx rpmrcCtxRelease(rpmrcCtx ctx)
+{
+    pthread_rwlock_unlock(&ctx->lock);
+    return NULL;
+}
 
 static int optionCompare(const void * a, const void * b)
 {
@@ -429,6 +465,7 @@ static void setDefaults(void)
 #ifndef MACROFILES
     if (!macrofiles) {
 	macrofiles = rstrscat(NULL, confdir, "/macros", ":",
+				confdir, "/macros.d/macros.*", ":",
 				confdir, "/platform/%{_target}/macros", ":",
 				confdir, "/fileattrs/*.attr", ":",
   				confdir, "/" RPMCANONVENDOR "/macros", ":",
@@ -443,7 +480,7 @@ static void setDefaults(void)
 }
 
 /* FIX: se usage inconsistent, W2DO? */
-static rpmRC doReadRC(const char * urlfn)
+static rpmRC doReadRC(rpmrcCtx ctx, const char * urlfn)
 {
     char *s;
     char *se, *next, *buf = NULL, *fn;
@@ -505,26 +542,18 @@ static rpmRC doReadRC(const char * urlfn)
 		goto exit;
 	    }
 
-	    switch (option->var) {
-	    case RPMVAR_INCLUDE:
+	    if (option->var == RPMVAR_INCLUDE) {
 		s = se;
 		while (*se && !risspace(*se)) se++;
-		if (*se != '\0') *se++ = '\0';
+		if (*se != '\0') *se = '\0';
 
-#if 0 /* XXX doesn't seem to do anything useful, only break things... */
-		rpmRebuildTargetVars(NULL, NULL);
-#endif
-
-		if (doReadRC(s)) {
+		if (doReadRC(ctx, s)) {
 		    rpmlog(RPMLOG_ERR, _("cannot open %s at %s:%d: %m\n"),
 			s, fn, linenum);
 		    goto exit;
-		} else {
-		    continue;	/* XXX don't save include value as var/macro */
 		}
-	      	break;
-	    default:
-		break;
+		/* XXX don't save include value as var/macro */
+		continue;
 	    }
 
 	    if (option->archSpecific) {
@@ -550,7 +579,7 @@ static rpmRC doReadRC(const char * urlfn)
 
 	    /* Only add macros if appropriate for this arch */
 	    if (option->macroize &&
-	      (arch == NULL || rstreq(arch, current[ARCH]))) {
+	      (arch == NULL || rstreq(arch, ctx->current[ARCH]))) {
 		char *n, *name;
 		n = name = xmalloc(strlen(option->name)+2);
 		if (option->localize)
@@ -559,7 +588,7 @@ static rpmRC doReadRC(const char * urlfn)
 		addMacro(NULL, name, NULL, val, RMIL_RPMRC);
 		free(name);
 	    }
-	    rpmSetVarArch(option->var, val, arch);
+	    rpmSetVarArch(ctx, option->var, val, arch);
 	    fn = _free(fn);
 
 	} else {	/* For arch/os compatibilty tables ... */
@@ -569,29 +598,30 @@ static rpmRC doReadRC(const char * urlfn)
 	    gotit = 0;
 
 	    for (i = 0; i < RPM_MACHTABLE_COUNT; i++) {
-		if (rstreqn(tables[i].key, s, strlen(tables[i].key)))
+		if (rstreqn(ctx->tables[i].key, s, strlen(ctx->tables[i].key)))
 		    break;
 	    }
 
 	    if (i < RPM_MACHTABLE_COUNT) {
-		const char *rest = s + strlen(tables[i].key);
+		const char *rest = s + strlen(ctx->tables[i].key);
 		if (*rest == '_') rest++;
 
 		if (rstreq(rest, "compat")) {
 		    if (machCompatCacheAdd(se, fn, linenum,
-						&tables[i].cache))
+						&ctx->tables[i].cache))
 			goto exit;
 		    gotit = 1;
-		} else if (tables[i].hasTranslate &&
+		} else if (ctx->tables[i].hasTranslate &&
 			   rstreq(rest, "translate")) {
-		    if (addDefault(&tables[i].defaults,
-				   &tables[i].defaultsLength,
+		    if (addDefault(&ctx->tables[i].defaults,
+				   &ctx->tables[i].defaultsLength,
 				   se, fn, linenum))
 			goto exit;
 		    gotit = 1;
-		} else if (tables[i].hasCanon &&
+		} else if (ctx->tables[i].hasCanon &&
 			   rstreq(rest, "canon")) {
-		    if (addCanon(&tables[i].canons, &tables[i].canonsLength,
+		    if (addCanon(&ctx->tables[i].canons,
+				 &ctx->tables[i].canonsLength,
 				 se, fn, linenum))
 			goto exit;
 		    gotit = 1;
@@ -617,7 +647,7 @@ exit:
 
 /**
  */
-static rpmRC rpmPlatform(const char * platform)
+static rpmRC rpmPlatform(rpmrcCtx ctx, const char * platform)
 {
     const char *cpu = NULL, *vendor = NULL, *os = NULL, *gnu = NULL;
     uint8_t * b = NULL;
@@ -650,10 +680,7 @@ static rpmRC rpmPlatform(const char * platform)
 	    while (--t > p && isspace(*t))
 		*t = '\0';
 	    if (t > p) {
-		platpat = xrealloc(platpat, (nplatpat + 2) * sizeof(*platpat));
-		platpat[nplatpat] = xstrdup(p);
-		nplatpat++;
-		platpat[nplatpat] = NULL;
+		argvAdd(&ctx->platpat, p);
 	    }
 	    continue;
 	}
@@ -670,7 +697,7 @@ static rpmRC rpmPlatform(const char * platform)
 	while (*p && !(*p == '-' || isspace(*p)))
 	    p++;
 	if (*p != '-') {
-	    if (*p != '\0') *p++ = '\0';
+	    if (*p != '\0') *p = '\0';
 	    os = vendor;
 	    vendor = "unknown";
 	} else {
@@ -686,17 +713,17 @@ static rpmRC rpmPlatform(const char * platform)
 		while (*p && !(*p == '-' || isspace(*p)))
 		    p++;
 	    }
-	    if (*p != '\0') *p++ = '\0';
+	    if (*p != '\0') *p = '\0';
 	}
 
 	addMacro(NULL, "_host_cpu", NULL, cpu, -1);
 	addMacro(NULL, "_host_vendor", NULL, vendor, -1);
 	addMacro(NULL, "_host_os", NULL, os, -1);
 
-	platpat = xrealloc(platpat, (nplatpat + 2) * sizeof(*platpat));
-	platpat[nplatpat] = rpmExpand("%{_host_cpu}-%{_host_vendor}-%{_host_os}", (gnu && *gnu ? "-" : NULL), gnu, NULL);
-	nplatpat++;
-	platpat[nplatpat] = NULL;
+	char *plat = rpmExpand("%{_host_cpu}-%{_host_vendor}-%{_host_os}",
+				(gnu && *gnu ? "-" : NULL), gnu, NULL);
+	argvAdd(&ctx->platpat, plat);
+	free(plat);
 	
 	init_platform++;
     }
@@ -812,8 +839,6 @@ static int is_athlon(void)
 	
 	cpuid (0, &eax, &ebx, &ecx, &edx);
 
- 	/* If you care about space, you can just check ebx, ecx and edx directly
- 	   instead of forming a string first and then doing a strcmp */
  	memset(vendor, 0, sizeof(vendor));
  	
  	for (i=0; i<4; i++)
@@ -829,7 +854,7 @@ static int is_athlon(void)
 	return 1;
 }
 
-static int is_pentium3()
+static int is_pentium3(void)
 {
     unsigned int eax, ebx, ecx, edx, family, model;
     char vendor[16];
@@ -849,18 +874,6 @@ static int is_pentium3()
 	    case 7:	// Pentium III, Pentium III Xeon (model 7)
 	    case 8:	// Pentium III, Pentium III Xeon, Celeron (model 8)
 	    case 9:	// Pentium M
-			/*
-			    Intel recently announced its new technology for mobile platforms,
-			    named Centrino, and presents it as a big advance in mobile PCs.
-			    One of the main part of Centrino consists in a brand new CPU,
-			    the Pentium M, codenamed Banias, that we'll study in this review.
-			    A particularity of this CPU is that it was designed for mobile platform
-			    exclusively, unlike previous mobile CPU (Pentium III-M, Pentium 4-M)
-			    that share the same micro-architecture as their desktop counterparts.
-			    The Pentium M introduces a new micro-architecture, adapted for mobility
-			    constraints, and that is halfway between the Pentium III and the Pentium 4.
-						    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
-			*/
 	    case 10:	// Pentium III Xeon (model A)
 	    case 11:	// Pentium III (model B)
 		return 1;
@@ -868,7 +881,7 @@ static int is_pentium3()
     return 0;
 }
 
-static int is_pentium4()
+static int is_pentium4(void)
 {
     unsigned int eax, ebx, ecx, edx, family, model;
     char vendor[16];
@@ -896,12 +909,10 @@ static int is_pentium4()
     return 0;
 }
 
-static int is_geode()
+static int is_geode(void)
 {
     unsigned int eax, ebx, ecx, edx, family, model;
     char vendor[16];
-    /* If you care about space, you can just check ebx, ecx and edx directly
-       instead of forming a string first and then doing a strcmp */
     memset(vendor, 0, sizeof(vendor));
     
     cpuid(0, &eax, &ebx, &ecx, &edx);
@@ -924,38 +935,80 @@ static int is_geode()
 }
 #endif
 
-#if defined(__linux__) && defined(__powerpc__)
-static jmp_buf mfspr_jmpbuf;
 
-static void mfspr_ill(int notused)
+#if defined(__linux__)
+/**
+ * Populate rpmat structure with auxv values
+ */
+static void read_auxv(void)
 {
-    longjmp(mfspr_jmpbuf, -1);
+    static int oneshot = 1;
+
+    if (oneshot) {
+#ifdef HAVE_GETAUXVAL
+	rpmat.platform = (char *) getauxval(AT_PLATFORM);
+	if (!rpmat.platform)
+	    rpmat.platform = "";
+	rpmat.hwcap = getauxval(AT_HWCAP);
+#else
+	rpmat.platform = "";
+	int fd = open("/proc/self/auxv", O_RDONLY);
+
+	if (fd == -1) {
+	    rpmlog(RPMLOG_WARNING,
+		   _("Failed to read auxiliary vector, /proc not mounted?\n"));
+            return;
+	} else {
+	    ElfW(auxv_t) auxv;
+	    while (read(fd, &auxv, sizeof(auxv)) == sizeof(auxv)) {
+                switch (auxv.a_type)
+                {
+                    case AT_NULL:
+                        break;
+                    case AT_PLATFORM:
+                        rpmat.platform = strdup((char *) auxv.a_un.a_val);
+                        break;
+                    case AT_HWCAP:
+                        rpmat.hwcap = auxv.a_un.a_val;
+                        break;
+                }
+	    }
+	    close(fd);
+	}
+#endif
+	oneshot = 0; /* only try once even if it fails */
+    }
+    return;
 }
 #endif
 
 /**
  */
-static void defaultMachine(const char ** arch,
-		const char ** os)
+static void defaultMachine(rpmrcCtx ctx, const char ** arch, const char ** os)
 {
+    const char * const platform_path = SYSCONFDIR "/rpm/platform";
     static struct utsname un;
     static int gotDefaults = 0;
     char * chptr;
     canonEntry canon;
     int rc;
 
+#if defined(__linux__)
+    /* Populate rpmat struct with hw info */
+    read_auxv();
+#endif
+
     while (!gotDefaults) {
-	if (!rpmPlatform(platform)) {
-	    char * s;
-	    s = rpmExpand("%{_host_cpu}", NULL);
+	if (!rpmPlatform(ctx, platform_path)) {
+	    char * s = rpmExpand("%{_host_cpu}", NULL);
 	    if (s) {
 		rstrlcpy(un.machine, s, sizeof(un.machine));
-		s = _free(s);
+		free(s);
 	    }
 	    s = rpmExpand("%{_host_os}", NULL);
 	    if (s) {
 		rstrlcpy(un.sysname, s, sizeof(un.sysname));
-		s = _free(s);
+		free(s);
 	    }
 	    gotDefaults = 1;
 	    break;
@@ -964,12 +1017,6 @@ static void defaultMachine(const char ** arch,
 	if (rc < 0) return;
 
 #if !defined(__linux__)
-#ifdef SNI
-	/* USUALLY un.sysname on sinix does start with the word "SINIX"
-	 * let's be absolutely sure
-	 */
-	strncpy(un.sysname, "SINIX", sizeof(un.sysname));
-#endif
 	if (rstreq(un.sysname, "AIX")) {
 	    strcpy(un.machine, __power_pc() ? "ppc" : "rs6000");
 	    sprintf(un.sysname,"aix%s.%s", un.version, un.release);
@@ -982,22 +1029,9 @@ static void defaultMachine(const char ** arch,
 #endif 
 	}
 	else if (rstreq(un.sysname, "SunOS")) {
-	    if (rstreqn(un.release,"4", 1)) /* SunOS 4.x */ {
-		int fd;
-		for (fd = 0;
-		    (un.release[fd] != 0 && (fd < sizeof(un.release)));
-		    fd++) {
-		      if (!risdigit(un.release[fd]) && (un.release[fd] != '.')) {
-			un.release[fd] = 0;
-			break;
-		      }
-		    }
-		    sprintf(un.sysname,"sunos%s",un.release);
-	    }
-
-	    else /* Solaris 2.x: n.x.x becomes n-3.x.x */
-		sprintf(un.sysname, "solaris%1d%s", atoi(un.release)-3,
-			un.release+1+(atoi(un.release)/10));
+	    /* Solaris 2.x: n.x.x becomes n-3.x.x */
+	    sprintf(un.sysname, "solaris%1d%s", atoi(un.release)-3,
+		    un.release+1+(atoi(un.release)/10));
 
 	    /* Solaris on Intel hardware reports i86pc instead of i386
 	     * (at least on 2.6 and 2.8)
@@ -1011,41 +1045,6 @@ static void defaultMachine(const char ** arch,
 	else if (rstreq(un.sysname, "OSF1"))
 	    /*make un.sysname look like osf3.2 for example*/
 	    sprintf(un.sysname, "osf%s", strpbrk(un.release, "123456789"));
-	else if (rstreqn(un.sysname, "IP", 2))
-	    un.sysname[2] = '\0';
-	else if (rstreqn(un.sysname, "SINIX", 5)) {
-	    sprintf(un.sysname, "sinix%s",un.release);
-	    if (rstreqn(un.machine, "RM", 2))
-		sprintf(un.machine, "mips");
-	}
-	else if ((rstreqn(un.machine, "34", 2) ||
-		rstreqn(un.machine, "33", 2)) && \
-		rstreqn(un.release, "4.0", 3))
-	{
-	    /* we are on ncr-sysv4 */
-	    char * prelid = NULL;
-	    FD_t fd = Fopen("/etc/.relid", "r.fdio");
-	    int gotit = 0;
-	    if (fd != NULL && !Ferror(fd)) {
-		chptr = xcalloc(1, 256);
-		{   int irelid = Fread(chptr, sizeof(*chptr), 256, fd);
-		    (void) Fclose(fd);
-		    /* example: "112393 RELEASE 020200 Version 01 OS" */
-		    if (irelid > 0) {
-			if ((prelid = strstr(chptr, "RELEASE "))){
-			    prelid += strlen("RELEASE ")+1;
-			    sprintf(un.sysname,"ncr-sysv4.%.*s",1,prelid);
-			    gotit = 1;
-			}
-		    }
-		}
-		chptr = _free (chptr);
-	    }
-	    if (!gotit)	/* parsing /etc/.relid file failed? */
-		strcpy(un.sysname,"ncr-sysv4");
-	    /* wrong, just for now, find out how to look for i586 later*/
-	    strcpy(un.machine,"i486");
-	}
 #endif	/* __linux__ */
 
 	/* get rid of the hyphens in the sysname */
@@ -1103,6 +1102,9 @@ static void defaultMachine(const char ** arch,
 #	endif	/* hpux */
 
 #	if defined(__linux__) && defined(__sparc__)
+#	if !defined(HWCAP_SPARC_BLKINIT)
+#	    define HWCAP_SPARC_BLKINIT	0x00000040
+#	endif
 	if (rstreq(un.machine, "sparc")) {
 	    #define PERS_LINUX		0x00000000
 	    #define PERS_LINUX_32BIT	0x00800000
@@ -1122,9 +1124,54 @@ static void defaultMachine(const char ** arch,
 		}
 		personality(oldpers);
 	    }
+
+	    /* This is how glibc detects Niagara so... */
+	    if (rpmat.hwcap & HWCAP_SPARC_BLKINIT) {
+		if (rstreq(un.machine, "sparcv9") || rstreq(un.machine, "sparc")) {
+		    strcpy(un.machine, "sparcv9v");
+		} else if (rstreq(un.machine, "sparc64")) {
+		    strcpy(un.machine, "sparc64v");
+		}
+	    }
 	}
 #	endif	/* sparc*-linux */
 
+#	if defined(__linux__) && defined(__powerpc__)
+#	if __BYTE_ORDER__ == __ORDER_BIG_ENDIAN__
+	{
+            int powerlvl;
+            if (!rstreq(un.machine, "ppc") &&
+		    sscanf(rpmat.platform, "power%d", &powerlvl) == 1 &&
+		    powerlvl > 6) {
+                strcpy(un.machine, "ppc64p7");
+	    }
+        }
+#	endif	/* __ORDER_BIG_ENDIAN__ */
+#	endif	/* ppc64*-linux */
+
+#	if defined(__linux__) && defined(__arm__) && defined(__ARM_PCS_VFP)
+#	if !defined(HWCAP_ARM_VFP)
+#	    define HWCAP_ARM_VFP	(1 << 6)
+#	endif
+#	if !defined(HWCAP_ARM_NEON)
+#	    define HWCAP_ARM_NEON	(1 << 12)
+#	endif
+#	if !defined(HWCAP_ARM_VFPv3)
+#	    define HWCAP_ARM_VFPv3	(1 << 13)
+#	endif
+	if (rstreq(un.machine, "armv7l")) {
+	    if (rpmat.hwcap & HWCAP_ARM_VFPv3) {
+		if (rpmat.hwcap & HWCAP_ARM_NEON)
+		    strcpy(un.machine, "armv7hnl");
+		else
+		    strcpy(un.machine, "armv7hl");
+	    }
+	} else if (rstreq(un.machine, "armv6l")) {
+	    if (rpmat.hwcap & HWCAP_ARM_VFP)
+		strcpy(un.machine, "armv6hl");
+	}
+#	endif	/* arm*-linux */
+
 #	if defined(__GNUC__) && defined(__alpha__)
 	{
 	    unsigned long amask, implver;
@@ -1170,14 +1217,14 @@ static void defaultMachine(const char ** arch,
 
 	/* the uname() result goes through the arch_canon table */
 	canon = lookupInCanonTable(un.machine,
-				   tables[RPM_MACHTABLE_INSTARCH].canons,
-				   tables[RPM_MACHTABLE_INSTARCH].canonsLength);
+			   ctx->tables[RPM_MACHTABLE_INSTARCH].canons,
+			   ctx->tables[RPM_MACHTABLE_INSTARCH].canonsLength);
 	if (canon)
 	    rstrlcpy(un.machine, canon->short_name, sizeof(un.machine));
 
 	canon = lookupInCanonTable(un.sysname,
-				   tables[RPM_MACHTABLE_INSTOS].canons,
-				   tables[RPM_MACHTABLE_INSTOS].canonsLength);
+			   ctx->tables[RPM_MACHTABLE_INSTOS].canons,
+			   ctx->tables[RPM_MACHTABLE_INSTOS].canonsLength);
 	if (canon)
 	    rstrlcpy(un.sysname, canon->short_name, sizeof(un.sysname));
 	gotDefaults = 1;
@@ -1189,34 +1236,30 @@ static void defaultMachine(const char ** arch,
 }
 
 static
-const char * rpmGetVarArch(int var, const char * arch)
+const char * rpmGetVarArch(rpmrcCtx ctx, int var, const char * arch)
 {
     const struct rpmvarValue * next;
 
-    if (arch == NULL) arch = current[ARCH];
+    if (arch == NULL) arch = ctx->current[ARCH];
 
     if (arch) {
-	next = &values[var];
+	next = &ctx->values[var];
 	while (next) {
 	    if (next->arch && rstreq(next->arch, arch)) return next->value;
 	    next = next->next;
 	}
     }
 
-    next = values + var;
+    next = ctx->values + var;
     while (next && next->arch) next = next->next;
 
     return next ? next->value : NULL;
 }
 
-static const char *rpmGetVar(int var)
+static void rpmSetVarArch(rpmrcCtx ctx,
+			  int var, const char * val, const char * arch)
 {
-    return rpmGetVarArch(var, NULL);
-}
-
-static void rpmSetVarArch(int var, const char * val, const char * arch)
-{
-    struct rpmvarValue * next = values + var;
+    struct rpmvarValue * next = ctx->values + var;
 
     if (next->value) {
 	if (arch) {
@@ -1248,39 +1291,23 @@ static void rpmSetVarArch(int var, const char * val, const char * arch)
     next->arch = (arch ? xstrdup(arch) : NULL);
 }
 
-void rpmSetTables(int archTable, int osTable)
+static void rpmSetTables(rpmrcCtx ctx, int archTable, int osTable)
 {
     const char * arch, * os;
 
-    defaultMachine(&arch, &os);
+    defaultMachine(ctx, &arch, &os);
 
-    if (currTables[ARCH] != archTable) {
-	currTables[ARCH] = archTable;
-	rebuildCompatTables(ARCH, arch);
+    if (ctx->currTables[ARCH] != archTable) {
+	ctx->currTables[ARCH] = archTable;
+	rebuildCompatTables(ctx, ARCH, arch);
     }
 
-    if (currTables[OS] != osTable) {
-	currTables[OS] = osTable;
-	rebuildCompatTables(OS, os);
+    if (ctx->currTables[OS] != osTable) {
+	ctx->currTables[OS] = osTable;
+	rebuildCompatTables(ctx, OS, os);
     }
 }
 
-int rpmMachineScore(int type, const char * name)
-{
-    machEquivInfo info = NULL;
-    if (name)
-	info = machEquivSearch(&tables[type].equiv, name);
-    return info ? info->score : 0;
-}
-
-int rpmIsKnownArch(const char *name)
-{
-    canonEntry canon = lookupInCanonTable(name,
-			tables[RPM_MACHTABLE_INSTARCH].canons,
-			tables[RPM_MACHTABLE_INSTARCH].canonsLength);
-    return (canon != NULL || rstreq(name, "noarch"));
-}
-
 /** \ingroup rpmrc
  * Set current arch/os names.
  * NULL as argument is set to the default value (munged uname())
@@ -1288,43 +1315,44 @@ int rpmIsKnownArch(const char *name)
  * @deprecated Use addMacro to set _target_* macros.
  * @todo Eliminate 
  *
+ * @param ctx		rpmrc context
  * @param arch          arch name (or NULL)
  * @param os            os name (or NULL)
  *          */
 
-static void rpmSetMachine(const char * arch, const char * os)
+static void rpmSetMachine(rpmrcCtx ctx, const char * arch, const char * os)
 {
     const char * host_cpu, * host_os;
 
-    defaultMachine(&host_cpu, &host_os);
+    defaultMachine(ctx, &host_cpu, &host_os);
 
     if (arch == NULL) {
 	arch = host_cpu;
-	if (tables[currTables[ARCH]].hasTranslate)
+	if (ctx->tables[ctx->currTables[ARCH]].hasTranslate)
 	    arch = lookupInDefaultTable(arch,
-			    tables[currTables[ARCH]].defaults,
-			    tables[currTables[ARCH]].defaultsLength);
+			    ctx->tables[ctx->currTables[ARCH]].defaults,
+			    ctx->tables[ctx->currTables[ARCH]].defaultsLength);
     }
     if (arch == NULL) return;	/* XXX can't happen */
 
     if (os == NULL) {
 	os = host_os;
-	if (tables[currTables[OS]].hasTranslate)
+	if (ctx->tables[ctx->currTables[OS]].hasTranslate)
 	    os = lookupInDefaultTable(os,
-			    tables[currTables[OS]].defaults,
-			    tables[currTables[OS]].defaultsLength);
+			    ctx->tables[ctx->currTables[OS]].defaults,
+			    ctx->tables[ctx->currTables[OS]].defaultsLength);
     }
     if (os == NULL) return;	/* XXX can't happen */
 
-    if (!current[ARCH] || !rstreq(arch, current[ARCH])) {
-	current[ARCH] = _free(current[ARCH]);
-	current[ARCH] = xstrdup(arch);
-	rebuildCompatTables(ARCH, host_cpu);
+    if (!ctx->current[ARCH] || !rstreq(arch, ctx->current[ARCH])) {
+	ctx->current[ARCH] = _free(ctx->current[ARCH]);
+	ctx->current[ARCH] = xstrdup(arch);
+	rebuildCompatTables(ctx, ARCH, host_cpu);
     }
 
-    if (!current[OS] || !rstreq(os, current[OS])) {
+    if (!ctx->current[OS] || !rstreq(os, ctx->current[OS])) {
 	char * t = xstrdup(os);
-	current[OS] = _free(current[OS]);
+	ctx->current[OS] = _free(ctx->current[OS]);
 	/*
 	 * XXX Capitalizing the 'L' is needed to insure that old
 	 * XXX os-from-uname (e.g. "Linux") is compatible with the new
@@ -1335,57 +1363,49 @@ static void rpmSetMachine(const char * arch, const char * os)
 	 */
 	if (rstreq(t, "linux"))
 	    *t = 'L';
-	current[OS] = t;
+	ctx->current[OS] = t;
 	
-	rebuildCompatTables(OS, host_os);
+	rebuildCompatTables(ctx, OS, host_os);
     }
 }
 
-static void rebuildCompatTables(int type, const char * name)
+static void rebuildCompatTables(rpmrcCtx ctx, int type, const char * name)
 {
-    machFindEquivs(&tables[currTables[type]].cache,
-		   &tables[currTables[type]].equiv,
+    machFindEquivs(&ctx->tables[ctx->currTables[type]].cache,
+		   &ctx->tables[ctx->currTables[type]].equiv,
 		   name);
 }
 
-static void getMachineInfo(int type, const char ** name,
-			int * num)
+static void getMachineInfo(rpmrcCtx ctx,
+			   int type, const char ** name, int * num)
 {
     canonEntry canon;
-    int which = currTables[type];
+    int which = ctx->currTables[type];
 
     /* use the normal canon tables, even if we're looking up build stuff */
     if (which >= 2) which -= 2;
 
-    canon = lookupInCanonTable(current[type],
-			       tables[which].canons,
-			       tables[which].canonsLength);
+    canon = lookupInCanonTable(ctx->current[type],
+			       ctx->tables[which].canons,
+			       ctx->tables[which].canonsLength);
 
     if (canon) {
 	if (num) *num = canon->num;
 	if (name) *name = canon->short_name;
     } else {
 	if (num) *num = 255;
-	if (name) *name = current[type];
+	if (name) *name = ctx->current[type];
 
-	if (tables[currTables[type]].hasCanon) {
-	    rpmlog(RPMLOG_WARNING, _("Unknown system: %s\n"), current[type]);
+	if (ctx->tables[ctx->currTables[type]].hasCanon) {
+	    rpmlog(RPMLOG_WARNING, _("Unknown system: %s\n"),
+		   ctx->current[type]);
 	    rpmlog(RPMLOG_WARNING, _("Please contact %s\n"), PACKAGE_BUGREPORT);
 	}
     }
 }
 
-void rpmGetArchInfo(const char ** name, int * num)
-{
-    getMachineInfo(ARCH, name, num);
-}
-
-void rpmGetOsInfo(const char ** name, int * num)
-{
-    getMachineInfo(OS, name, num);
-}
-
-static void rpmRebuildTargetVars(const char ** target, const char ** canontarget)
+static void rpmRebuildTargetVars(rpmrcCtx ctx,
+			const char ** target, const char ** canontarget)
 {
 
     char *ca = NULL, *co = NULL, *ct = NULL;
@@ -1393,9 +1413,9 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
 
     /* Rebuild the compat table to recalculate the current target arch.  */
 
-    rpmSetMachine(NULL, NULL);
-    rpmSetTables(RPM_MACHTABLE_INSTARCH, RPM_MACHTABLE_INSTOS);
-    rpmSetTables(RPM_MACHTABLE_BUILDARCH, RPM_MACHTABLE_BUILDOS);
+    rpmSetMachine(ctx, NULL, NULL);
+    rpmSetTables(ctx, RPM_MACHTABLE_INSTARCH, RPM_MACHTABLE_INSTOS);
+    rpmSetTables(ctx, RPM_MACHTABLE_BUILDARCH, RPM_MACHTABLE_BUILDOS);
 
     if (target && *target) {
 	char *c;
@@ -1420,16 +1440,16 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
 	const char *a = NULL;
 	const char *o = NULL;
 	/* Set build target from rpm arch and os */
-	rpmGetArchInfo(&a, NULL);
+	getMachineInfo(ctx, ARCH, &a, NULL);
 	ca = (a) ? xstrdup(a) : NULL;
-	rpmGetOsInfo(&o, NULL);
+	getMachineInfo(ctx, OS, &o, NULL);
 	co = (o) ? xstrdup(o) : NULL;
     }
 
     /* If still not set, Set target arch/os from default uname(2) values */
     if (ca == NULL) {
 	const char *a = NULL;
-	defaultMachine(&a, NULL);
+	defaultMachine(ctx, &a, NULL);
 	ca = xstrdup(a ? a : "(arch)");
     }
     for (x = 0; ca[x] != '\0'; x++)
@@ -1437,7 +1457,7 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
 
     if (co == NULL) {
 	const char *o = NULL;
-	defaultMachine(NULL, &o);
+	defaultMachine(ctx, NULL, &o);
 	co = xstrdup(o ? o : "(os)");
     }
     for (x = 0; co[x] != '\0'; x++)
@@ -1461,7 +1481,7 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
 /*
  * XXX Make sure that per-arch optflags is initialized correctly.
  */
-  { const char *optflags = rpmGetVarArch(RPMVAR_OPTFLAGS, ca);
+  { const char *optflags = rpmGetVarArch(ctx, RPMVAR_OPTFLAGS, ca);
     if (optflags != NULL) {
 	delMacro(NULL, "optflags");
 	addMacro(NULL, "optflags", NULL, optflags, RMIL_RPMRC);
@@ -1471,24 +1491,122 @@ static void rpmRebuildTargetVars(const char ** target, const char ** canontarget
     if (canontarget)
 	*canontarget = ct;
     else
-	ct = _free(ct);
-    ca = _free(ca);
-    co = _free(co);
+	free(ct);
+    free(ca);
+    free(co);
+}
+
+/** \ingroup rpmrc
+ * Read rpmrc (and macro) configuration file(s).
+ * @param ctx		rpmrc context
+ * @param rcfiles	colon separated files to read (NULL uses default)
+ * @return		RPMRC_OK on success
+ */
+static rpmRC rpmReadRC(rpmrcCtx ctx, const char * rcfiles)
+{
+    ARGV_t p, globs = NULL, files = NULL;
+    rpmRC rc = RPMRC_FAIL;
+
+    if (!defaultsInitialized) {
+	setDefaults();
+	defaultsInitialized = 1;
+    }
+
+    if (rcfiles == NULL)
+	rcfiles = defrcfiles;
+
+    /* Expand any globs in rcfiles. Missing files are ok here. */
+    argvSplit(&globs, rcfiles, ":");
+    for (p = globs; *p; p++) {
+	ARGV_t av = NULL;
+	if (rpmGlob(*p, NULL, &av) == 0) {
+	    argvAppend(&files, av);
+	    argvFree(av);
+	}
+    }
+    argvFree(globs);
+
+    /* Read each file in rcfiles. */
+    for (p = files; p && *p; p++) {
+	/* XXX Only /usr/lib/rpm/rpmrc must exist in default rcfiles list */
+	if (access(*p, R_OK) != 0) {
+	    if (rcfiles == defrcfiles && p != files)
+		continue;
+	    rpmlog(RPMLOG_ERR, _("Unable to open %s for reading: %m.\n"), *p);
+	    goto exit;
+	    break;
+	} else {
+	    rc = doReadRC(ctx, *p);
+	}
+    }
+    rc = RPMRC_OK;
+    rpmSetMachine(ctx, NULL, NULL);	/* XXX WTFO? Why bother? */
+
+exit:
+    argvFree(files);
+    return rc;
+}
+
+/* External interfaces */
+
+int rpmReadConfigFiles(const char * file, const char * target)
+{
+    int rc = -1; /* assume failure */
+    rpmrcCtx ctx = rpmrcCtxAcquire(1);
+
+    /* Force preloading of dlopen()'ed libraries in case we go chrooting */
+    if (rpmugInit())
+	goto exit;
+
+    if (rpmInitCrypto())
+	goto exit;
+
+    /* Preset target macros */
+   	/* FIX: target can be NULL */
+    rpmRebuildTargetVars(ctx, &target, NULL);
+
+    /* Read the files */
+    if (rpmReadRC(ctx, file))
+	goto exit;
+
+    if (macrofiles != NULL) {
+	char *mf = rpmGetPath(macrofiles, NULL);
+	rpmInitMacros(NULL, mf);
+	_free(mf);
+    }
+
+    /* Reset target macros */
+    rpmRebuildTargetVars(ctx, &target, NULL);
+
+    /* Finally set target platform */
+    {	char *cpu = rpmExpand("%{_target_cpu}", NULL);
+	char *os = rpmExpand("%{_target_os}", NULL);
+	rpmSetMachine(ctx, cpu, os);
+	free(cpu);
+	free(os);
+    }
+
+#ifdef WITH_LUA
+    /* Force Lua state initialization */
+    rpmLuaInit();
+#endif
+    rc = 0;
+
+exit:
+    rpmrcCtxRelease(ctx);
+    return rc;
 }
 
 void rpmFreeRpmrc(void)
 {
+    rpmrcCtx ctx = rpmrcCtxAcquire(1);
     int i, j, k;
 
-    if (platpat)
-    for (i = 0; i < nplatpat; i++)
-	platpat[i] = _free(platpat[i]);
-    platpat = _free(platpat);
-    nplatpat = 0;
+    ctx->platpat = argvFree(ctx->platpat);
 
     for (i = 0; i < RPM_MACHTABLE_COUNT; i++) {
 	tableType t;
-	t = tables + i;
+	t = ctx->tables + i;
 	if (t->equiv.list) {
 	    for (j = 0; j < t->equiv.count; j++)
 		t->equiv.list[j].name = _free(t->equiv.list[j].name);
@@ -1531,19 +1649,18 @@ void rpmFreeRpmrc(void)
 
     for (i = 0; i < RPMVAR_NUM; i++) {
 	struct rpmvarValue * vp;
-	while ((vp = values[i].next) != NULL) {
-	    values[i].next = vp->next;
+	while ((vp = ctx->values[i].next) != NULL) {
+	    ctx->values[i].next = vp->next;
 	    vp->value = _free(vp->value);
 	    vp->arch = _free(vp->arch);
 	    vp = _free(vp);
 	}
-	values[i].value = _free(values[i].value);
-	values[i].arch = _free(values[i].arch);
+	ctx->values[i].value = _free(ctx->values[i].value);
+	ctx->values[i].arch = _free(ctx->values[i].arch);
     }
-    current[OS] = _free(current[OS]);
-    current[ARCH] = _free(current[ARCH]);
+    ctx->current[OS] = _free(ctx->current[OS]);
+    ctx->current[ARCH] = _free(ctx->current[ARCH]);
     defaultsInitialized = 0;
-/* FIX: platpat/current may be NULL */
 
     /* XXX doesn't really belong here but... */
     rpmFreeCrypto();
@@ -1551,150 +1668,65 @@ void rpmFreeRpmrc(void)
     rpmLuaFree();
 #endif
 
+    rpmrcCtxRelease(ctx);
     return;
 }
 
-/** \ingroup rpmrc
- * Read rpmrc (and macro) configuration file(s).
- * @param rcfiles	colon separated files to read (NULL uses default)
- * @return		RPMRC_OK on success
- */
-static rpmRC rpmReadRC(const char * rcfiles)
-{
-    ARGV_t p, globs = NULL, files = NULL;
-    rpmRC rc = RPMRC_FAIL;
-
-    if (!defaultsInitialized) {
-	setDefaults();
-	defaultsInitialized = 1;
-    }
-
-    if (rcfiles == NULL)
-	rcfiles = defrcfiles;
-
-    /* Expand any globs in rcfiles. Missing files are ok here. */
-    argvSplit(&globs, rcfiles, ":");
-    for (p = globs; *p; p++) {
-	ARGV_t av = NULL;
-	if (rpmGlob(*p, NULL, &av) == 0) {
-	    argvAppend(&files, av);
-	    argvFree(av);
-	}
-    }
-    argvFree(globs);
-
-    /* Read each file in rcfiles. */
-    for (p = files; p && *p; p++) {
-	/* XXX Only /usr/lib/rpm/rpmrc must exist in default rcfiles list */
-	if (access(*p, R_OK) != 0) {
-	    if (rcfiles == defrcfiles && p != files)
-		continue;
-	    rpmlog(RPMLOG_ERR, _("Unable to open %s for reading: %m.\n"), *p);
-	    goto exit;
-	    break;
-	} else {
-	    rc = doReadRC(*p);
-	}
-    }
-    rc = RPMRC_OK;
-    rpmSetMachine(NULL, NULL);	/* XXX WTFO? Why bother? */
-
-exit:
-    argvFree(files);
-    return rc;
-}
-
-int rpmReadConfigFiles(const char * file, const char * target)
-{
-    /* Force preloading of dlopen()'ed libraries in case we go chrooting */
-    (void) gethostbyname("localhost");
-    (void) rpmInitCrypto();
-
-    /* Preset target macros */
-   	/* FIX: target can be NULL */
-    rpmRebuildTargetVars(&target, NULL);
-
-    /* Read the files */
-    if (rpmReadRC(file)) return -1;
-
-    if (macrofiles != NULL) {
-	char *mf = rpmGetPath(macrofiles, NULL);
-	rpmInitMacros(NULL, mf);
-	_free(mf);
-    }
-
-    /* Reset target macros */
-    rpmRebuildTargetVars(&target, NULL);
-
-    /* Finally set target platform */
-    {	char *cpu = rpmExpand("%{_target_cpu}", NULL);
-	char *os = rpmExpand("%{_target_os}", NULL);
-	rpmSetMachine(cpu, os);
-	cpu = _free(cpu);
-	os = _free(os);
-    }
-
-#ifdef WITH_LUA
-    /* Force Lua state initialization */
-    rpmLuaInit();
-#endif
-
-    return 0;
-}
-
 int rpmShowRC(FILE * fp)
 {
+    /* Write-context necessary as this calls rpmSetTables(), ugh */
+    rpmrcCtx ctx = rpmrcCtxAcquire(1);
     const struct rpmOption *opt;
     rpmds ds = NULL;
-    int i, xx;
+    int i;
     machEquivTable equivTable;
 
     /* the caller may set the build arch which should be printed here */
     fprintf(fp, "ARCHITECTURE AND OS:\n");
-    fprintf(fp, "build arch            : %s\n", current[ARCH]);
+    fprintf(fp, "build arch            : %s\n", ctx->current[ARCH]);
 
     fprintf(fp, "compatible build archs:");
-    equivTable = &tables[RPM_MACHTABLE_BUILDARCH].equiv;
+    equivTable = &ctx->tables[RPM_MACHTABLE_BUILDARCH].equiv;
     for (i = 0; i < equivTable->count; i++)
 	fprintf(fp," %s", equivTable->list[i].name);
     fprintf(fp, "\n");
 
-    fprintf(fp, "build os              : %s\n", current[OS]);
+    fprintf(fp, "build os              : %s\n", ctx->current[OS]);
 
     fprintf(fp, "compatible build os's :");
-    equivTable = &tables[RPM_MACHTABLE_BUILDOS].equiv;
+    equivTable = &ctx->tables[RPM_MACHTABLE_BUILDOS].equiv;
     for (i = 0; i < equivTable->count; i++)
 	fprintf(fp," %s", equivTable->list[i].name);
     fprintf(fp, "\n");
 
-    rpmSetTables(RPM_MACHTABLE_INSTARCH, RPM_MACHTABLE_INSTOS);
-    rpmSetMachine(NULL, NULL);	/* XXX WTFO? Why bother? */
+    rpmSetTables(ctx, RPM_MACHTABLE_INSTARCH, RPM_MACHTABLE_INSTOS);
+    rpmSetMachine(ctx, NULL, NULL);	/* XXX WTFO? Why bother? */
 
-    fprintf(fp, "install arch          : %s\n", current[ARCH]);
-    fprintf(fp, "install os            : %s\n", current[OS]);
+    fprintf(fp, "install arch          : %s\n", ctx->current[ARCH]);
+    fprintf(fp, "install os            : %s\n", ctx->current[OS]);
 
     fprintf(fp, "compatible archs      :");
-    equivTable = &tables[RPM_MACHTABLE_INSTARCH].equiv;
+    equivTable = &ctx->tables[RPM_MACHTABLE_INSTARCH].equiv;
     for (i = 0; i < equivTable->count; i++)
 	fprintf(fp," %s", equivTable->list[i].name);
     fprintf(fp, "\n");
 
     fprintf(fp, "compatible os's       :");
-    equivTable = &tables[RPM_MACHTABLE_INSTOS].equiv;
+    equivTable = &ctx->tables[RPM_MACHTABLE_INSTOS].equiv;
     for (i = 0; i < equivTable->count; i++)
 	fprintf(fp," %s", equivTable->list[i].name);
     fprintf(fp, "\n");
 
     fprintf(fp, "\nRPMRC VALUES:\n");
     for (i = 0, opt = optionTable; i < optionTableSize; i++, opt++) {
-	const char *s = rpmGetVar(opt->var);
+	const char *s = rpmGetVarArch(ctx, opt->var, NULL);
 	if (s != NULL || rpmIsVerbose())
 	    fprintf(fp, "%-21s : %s\n", opt->name, s ? s : "(not set)");
     }
     fprintf(fp, "\n");
 
     fprintf(fp, "Features supported by rpmlib:\n");
-    xx = rpmdsRpmlib(&ds, NULL);
+    rpmdsRpmlib(&ds, NULL);
     ds = rpmdsInit(ds);
     while (rpmdsNext(ds) >= 0) {
         const char * DNEVR = rpmdsDNEVR(ds);
@@ -1704,7 +1736,74 @@ int rpmShowRC(FILE * fp)
     ds = rpmdsFree(ds);
     fprintf(fp, "\n");
 
+    fprintf(fp, "Macro path: %s\n", macrofiles);
+    fprintf(fp, "\n");
+
     rpmDumpMacroTable(NULL, fp);
 
+    /* XXX: Move this earlier eventually... */
+    rpmrcCtxRelease(ctx);
+
     return 0;
 }
+
+int rpmMachineScore(int type, const char * name)
+{
+    int score = 0;
+    if (name) {
+	rpmrcCtx ctx = rpmrcCtxAcquire(0);
+	machEquivInfo info = machEquivSearch(&ctx->tables[type].equiv, name);
+	if (info)
+	    score = info->score;
+	rpmrcCtxRelease(ctx);
+    }
+    return score;
+}
+
+int rpmIsKnownArch(const char *name)
+{
+    rpmrcCtx ctx = rpmrcCtxAcquire(0);
+    canonEntry canon = lookupInCanonTable(name,
+			    ctx->tables[RPM_MACHTABLE_INSTARCH].canons,
+			    ctx->tables[RPM_MACHTABLE_INSTARCH].canonsLength);
+    int known = (canon != NULL || rstreq(name, "noarch"));
+    rpmrcCtxRelease(ctx);
+    return known;
+}
+
+void rpmGetArchInfo(const char ** name, int * num)
+{
+    rpmrcCtx ctx = rpmrcCtxAcquire(0);
+    getMachineInfo(ctx, ARCH, name, num);
+    rpmrcCtxRelease(ctx);
+}
+
+int rpmGetArchColor(const char *arch)
+{
+    rpmrcCtx ctx = rpmrcCtxAcquire(0);
+    const char *color;
+    char *e;
+    int color_i = -1; /* assume failure */
+
+    arch = lookupInDefaultTable(arch,
+			    ctx->tables[ctx->currTables[ARCH]].defaults,
+			    ctx->tables[ctx->currTables[ARCH]].defaultsLength);
+    color = rpmGetVarArch(ctx, RPMVAR_ARCHCOLOR, arch);
+    if (color) {
+	color_i = strtol(color, &e, 10);
+	if (!(e && *e == '\0')) {
+	    color_i = -1;
+	}
+    }
+    rpmrcCtxRelease(ctx);
+
+    return color_i;
+}
+
+void rpmGetOsInfo(const char ** name, int * num)
+{
+    rpmrcCtx ctx = rpmrcCtxAcquire(0);
+    getMachineInfo(ctx, OS, name, num);
+    rpmrcCtxRelease(ctx);
+}
+
