@@ -44,26 +44,18 @@ rpmRC rpmReadSignature(FD_t fd, Header *sighp, sigType sig_type, char ** msg);
  */
 int rpmWriteSignature(FD_t fd, Header h);
 
-/** \ingroup signature
- * Generate digest(s) from a header+payload file, save in signature header.
- * @param sigh		signature header
- * @param file		header+payload file name
- * @param sigTag	type of digest(s) to add
- * @return		0 on success, -1 on failure
- */
-int rpmGenDigest(Header sigh, const char * file, rpmTagVal sigTag);
-
 /** \ingroup signature
  * Verify a signature from a package.
  *
  * @param keyring	keyring handle
  * @param sigtd		signature tag data container
- * @param dig		signature/pubkey parameters
+ * @param sig		signature/pubkey parameters
  * @retval result	detailed text result of signature verification
  * 			(malloc'd)
  * @return		result of signature verification
  */
-rpmRC rpmVerifySignature(rpmKeyring keyring, rpmtd sigtd, pgpDig dig, DIGEST_CTX ctx, char ** result);
+rpmRC rpmVerifySignature(rpmKeyring keyring, rpmtd sigtd, pgpDigParams sig,
+			 DIGEST_CTX ctx, char ** result);
 
 /** \ingroup signature
  * Destroy signature header from package.
@@ -72,6 +64,11 @@ rpmRC rpmVerifySignature(rpmKeyring keyring, rpmtd sigtd, pgpDig dig, DIGEST_CTX
  */
 Header rpmFreeSignature(Header h);
 
+/* Dumb wrapper around pgpPrtParams() to log some error messages on failure */
+RPM_GNUC_INTERNAL
+int parsePGPSig(rpmtd sigtd, const char *type,
+		 pgpDigParams *sig, char **msg);
+
 #ifdef __cplusplus
 }
 #endif
