@@ -33,21 +33,22 @@ struct headerTagFunc_s {
  * 
  * @param h		header
  * @param tagN		RPMTAG_BASENAMES | PMTAG_ORIGBASENAMES
- * @retval *fnp		array of file names
- * @retval *fcp		number of files
+ * @param withstate	take file state into account?
+ * @retval td		tag data container
+ * @return		1 on success
  */
-static void rpmfiBuildFNames(Header h, rpmTag tagN,
-	const char *** fnp, rpm_count_t * fcp)
+static int fnTag(Header h, rpmTag tagN, int withstate, rpmtd td)
 {
-    const char **baseNames, **dirNames, **fileNames;
+    const char **baseNames, **dirNames;
+    const char *fileStates = NULL;
     uint32_t *dirIndexes;
-    rpm_count_t count;
-    size_t size;
+    rpm_count_t count, retcount, dncount;
+    size_t size = 0;
     rpmTag dirNameTag = RPMTAG_DIRNAMES;
     rpmTag dirIndexesTag = RPMTAG_DIRINDEXES;
-    char * t;
-    int i;
-    struct rpmtd_s bnames, dnames, dixs;
+    int i, j;
+    int rc = 0; /* assume failure */
+    struct rpmtd_s bnames, dnames, dixs, fstates;
 
     if (tagN == RPMTAG_ORIGBASENAMES) {
 	dirNameTag = RPMTAG_ORIGDIRNAMES;
@@ -55,14 +56,31 @@ static void rpmfiBuildFNames(Header h, rpmTag tagN,
     }
 
     if (!headerGet(h, tagN, &bnames, HEADERGET_MINMEM)) {
-	*fnp = NULL;
-	*fcp = 0;
-	return;		/* no file list */
+	return 0;		/* no file list */
     }
+
     (void) headerGet(h, dirNameTag, &dnames, HEADERGET_MINMEM);
     (void) headerGet(h, dirIndexesTag, &dixs, HEADERGET_MINMEM);
 
-    count = rpmtdCount(&bnames);
+    retcount = count = rpmtdCount(&bnames);
+    dncount = rpmtdCount(&dnames);
+
+    /* Basic sanity checking for our interrelated tags  */
+    if (rpmtdCount(&dixs) != count || dncount < 1 || dncount > count)
+	td->flags |= RPMTD_INVALID;
+
+    if (withstate) {
+	/* no recorded states means no installed files */
+	if (!headerGet(h, RPMTAG_FILESTATES, &fstates, HEADERGET_MINMEM))
+	    goto exit;
+	if (rpmtdCount(&fstates) != count)
+	    td->flags |= RPMTD_INVALID;
+	fileStates = fstates.data;
+    }
+
+    if (td->flags & RPMTD_INVALID)
+	goto exit;
+
     baseNames = bnames.data;
     dirNames = dnames.data;
     dirIndexes = dixs.data;
@@ -72,23 +90,46 @@ static void rpmfiBuildFNames(Header h, rpmTag tagN,
      * block, until those assumptions are removed we need to jump through
      * a few hoops here and precalculate sizes etc
      */
-    size = sizeof(*fileNames) * count;
-    for (i = 0; i < count; i++)
-	size += strlen(baseNames[i]) + strlen(dirNames[dirIndexes[i]]) + 1;
-
-    fileNames = xmalloc(size);
-    t = ((char *) fileNames) + (sizeof(*fileNames) * count);
     for (i = 0; i < count; i++) {
-	fileNames[i] = t;
-	t = stpcpy( stpcpy(t, dirNames[dirIndexes[i]]), baseNames[i]);
-	*t++ = '\0';
+	if (fileStates && !RPMFILE_IS_INSTALLED(fileStates[i])) {
+	    retcount--;
+	    continue;
+	}
+	/* Sanity check  directory indexes are within bounds */
+	if (dirIndexes[i] >= dncount) {
+	    td->flags |= RPMTD_INVALID;
+	    break;
+	}
+	size += strlen(baseNames[i]) + strlen(dirNames[dirIndexes[i]]) + 1;
     }
+
+    if (!(td->flags & RPMTD_INVALID)) {
+	char **fileNames = xmalloc(size + (sizeof(*fileNames) * retcount));
+	char *t = ((char *) fileNames) + (sizeof(*fileNames) * retcount);
+	for (i = 0, j = 0; i < count; i++) {
+	    if (fileStates && !RPMFILE_IS_INSTALLED(fileStates[i]))
+		continue;
+	    fileNames[j++] = t;
+	    t = stpcpy( stpcpy(t, dirNames[dirIndexes[i]]), baseNames[i]);
+	    *t++ = '\0';
+	}
+
+	td->data = fileNames;
+	td->count = retcount;
+	td->type = RPM_STRING_ARRAY_TYPE;
+	td->flags |= RPMTD_ALLOCED;
+	rc = 1;
+    }
+
+exit:
     rpmtdFreeData(&bnames);
     rpmtdFreeData(&dnames);
     rpmtdFreeData(&dixs);
+    /* only safe if the headerGet() on file states was actually called */
+    if (fileStates)
+	rpmtdFreeData(&fstates);
 
-    *fnp = fileNames;
-    *fcp = count;
+    return rc;
 }
 
 static int filedepTag(Header h, rpmTag tagN, rpmtd td, headerGetFlags hgflags)
@@ -145,8 +186,8 @@ static int filedepTag(Header h, rpmTag tagN, rpmtd td, headerGetFlags hgflags)
     rc = 1;
 
 exit:
-    fi = rpmfiFree(fi);
-    ds = rpmdsFree(ds);
+    rpmfiFree(fi);
+    rpmdsFree(ds);
     return rc;
 }
 
@@ -273,6 +314,16 @@ static int triggertypeTag(Header h, rpmtd td, headerGetFlags hgflags)
     return 1;
 }
 
+/**
+ * Retrieve installed file paths.
+ * @param h		header
+ * @retval td		tag data container
+ * @return		1 on success
+ */
+static int instfilenamesTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return fnTag(h, RPMTAG_BASENAMES, 1, td);
+}
 /**
  * Retrieve file paths.
  * @param h		header
@@ -281,13 +332,7 @@ static int triggertypeTag(Header h, rpmtd td, headerGetFlags hgflags)
  */
 static int filenamesTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
-    rpmfiBuildFNames(h, RPMTAG_BASENAMES, 
-		     (const char ***) &(td->data), &(td->count));
-    if (td->data) {
-	td->type = RPM_STRING_ARRAY_TYPE;
-	td->flags = RPMTD_ALLOCED;
-    }
-    return (td->data != NULL); 
+    return fnTag(h, RPMTAG_BASENAMES, 0, td);
 }
 
 /**
@@ -298,16 +343,50 @@ static int filenamesTag(Header h, rpmtd td, headerGetFlags hgflags)
  */
 static int origfilenamesTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
-    rpmfiBuildFNames(h, RPMTAG_ORIGBASENAMES, 
-		     (const char ***) &(td->data), &(td->count));
-    if (td->data) {
-	td->type = RPM_STRING_ARRAY_TYPE;
-	td->flags = RPMTD_ALLOCED;
-    }
-    return (td->data != NULL); 
+    return fnTag(h, RPMTAG_ORIGBASENAMES, 0, td);
 }
+
+/*
+ * Attempt to generate libmagic-style file class if missing from header:
+ * we can easily generate this for symlinks and other special types.
+ * Always return malloced strings to simplify life in fileclassTag().
+ */
+static char *makeFClass(rpmfi fi)
+{
+    char *fclass = NULL;
+    const char *hc = rpmfiFClass(fi);
+
+    if (hc != NULL && hc[0] != '\0') {
+	fclass = xstrdup(hc);
+    } else {
+	switch (rpmfiFMode(fi) & S_IFMT) {
+	case S_IFBLK:
+	    fclass = xstrdup("block special");
+	    break;
+	case S_IFCHR:
+	    fclass = xstrdup("character special");
+	    break;
+	case S_IFDIR:
+	    fclass = xstrdup("directory");
+	    break;
+	case S_IFIFO:
+	    fclass = xstrdup("fifo (named pipe)");
+	    break;
+	case S_IFSOCK:
+	    fclass = xstrdup("socket");
+	    break;
+	case S_IFLNK:
+	    fclass = rstrscat(NULL, "symbolic link to `",
+			      rpmfiFLink(fi), "'", NULL);
+	    break;
+	}
+    }
+
+    return (fclass != NULL) ? fclass : xstrdup("");
+}
+
 /**
- * Retrieve file classes.
+ * Retrieve/generate file classes.
  * @param h		header
  * @retval td		tag data container
  * @return		1 on success
@@ -315,31 +394,25 @@ static int origfilenamesTag(Header h, rpmtd td, headerGetFlags hgflags)
 static int fileclassTag(Header h, rpmtd td, headerGetFlags hgflags)
 {
     rpmfi fi = rpmfiNew(NULL, h, RPMTAG_BASENAMES, RPMFI_NOHEADER);
-    char **fclasses;
-    int ix, numfiles;
-    int rc = 0;
+    int numfiles = rpmfiFC(fi);
 
-    numfiles = rpmfiFC(fi);
-    if (numfiles <= 0) {
-	goto exit;
+    if (numfiles > 0) {
+	char **fclasses = xmalloc(numfiles * sizeof(*fclasses));
+	int ix;
+
+	rpmfiInit(fi, 0);
+	while ((ix = rpmfiNext(fi)) >= 0) {
+	    fclasses[ix] = makeFClass(fi);
+	}
+
+	td->data = fclasses;
+	td->count = numfiles;
+	td->flags = RPMTD_ALLOCED | RPMTD_PTR_ALLOCED;
+	td->type = RPM_STRING_ARRAY_TYPE;
     }
 
-    fclasses = xmalloc(numfiles * sizeof(*fclasses));
-    rpmfiInit(fi, 0);
-    while ((ix = rpmfiNext(fi)) >= 0) {
-	const char *fclass = rpmfiFClass(fi);
-	fclasses[ix] = xstrdup(fclass ? fclass : "");
-    }
-
-    td->data = fclasses;
-    td->count = numfiles;
-    td->flags = RPMTD_ALLOCED | RPMTD_PTR_ALLOCED;
-    td->type = RPM_STRING_ARRAY_TYPE;
-    rc = 1;
-
-exit:
-    fi = rpmfiFree(fi);
-    return rc; 
+    rpmfiFree(fi);
+    return (numfiles > 0); 
 }
 
 /**
@@ -431,7 +504,7 @@ static int i18nTag(Header h, rpmTag tag, rpmtd td, headerGetFlags hgflags)
 	    return 1;
     }
 
-    dstring = _free(dstring);
+    free(dstring);
 #endif
 
     rc = headerGet(h, tag, td, HEADERGET_ALLOC);
@@ -662,6 +735,120 @@ static int epochnumTag(Header h, rpmtd td, headerGetFlags hgflags)
     return 1;
 }
 
+static int depnevrsTag(Header h, rpmtd td, headerGetFlags hgflags,
+			rpmTagVal tag)
+{
+    rpmds ds = rpmdsNew(h, tag, 0);
+    int ndeps = rpmdsCount(ds);
+
+    if (ndeps > 0) {
+	char **deps = xmalloc(sizeof(*deps) * ndeps);
+	int i;
+	while ((i = rpmdsNext(ds)) >= 0) {
+	    deps[i] = rpmdsNewDNEVR(NULL, ds);
+	}
+	td->data = deps;
+	td->type = RPM_STRING_ARRAY_TYPE;
+	td->count = ndeps;
+	td->flags |= (RPMTD_ALLOCED | RPMTD_PTR_ALLOCED);
+    }
+    rpmdsFree(ds);
+    return (ndeps > 0);
+}
+
+#define RPMSENSE_STRONG (1 << 27)
+
+static int depnevrsTagFiltered(Header h, rpmtd td, headerGetFlags hgflags,
+			rpmTagVal tag, int strong)
+{
+    rpmds ds = rpmdsNew(h, tag, 0);
+    int ndeps = rpmdsCount(ds);
+
+    if (ndeps > 0) {
+	char **deps = xmalloc(sizeof(*deps) * ndeps);
+	ndeps = 0;
+	while (rpmdsNext(ds) >= 0) {
+	    if ((rpmdsFlags(ds) & RPMSENSE_STRONG) == (strong ? RPMSENSE_STRONG : 0))
+		deps[ndeps++] = rpmdsNewDNEVR(NULL, ds);
+	}
+	if (ndeps) {
+	    td->data = deps;
+	    td->type = RPM_STRING_ARRAY_TYPE;
+	    td->count = ndeps;
+	    td->flags |= (RPMTD_ALLOCED | RPMTD_PTR_ALLOCED);
+	} else {
+	    _free(deps);
+        }
+    }
+    rpmdsFree(ds);
+    return (ndeps > 0);
+}
+
+static int requirenevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_REQUIRENAME);
+}
+
+static int recommendnevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_RECOMMENDNAME) ||
+           depnevrsTagFiltered(h, td, hgflags, RPMTAG_OLDSUGGESTSNAME, 1);
+}
+
+static int suggestnevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_SUGGESTNAME) ||
+           depnevrsTagFiltered(h, td, hgflags, RPMTAG_OLDSUGGESTSNAME, 0);
+}
+
+static int supplementnevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_SUPPLEMENTNAME) ||
+           depnevrsTagFiltered(h, td, hgflags, RPMTAG_OLDENHANCESNAME, 1);
+}
+
+static int enhancenevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_ENHANCENAME) ||
+           depnevrsTagFiltered(h, td, hgflags, RPMTAG_OLDENHANCESNAME, 0);
+}
+
+static int providenevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_PROVIDENAME);
+}
+
+static int obsoletenevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_OBSOLETENAME);
+}
+
+static int conflictnevrsTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    return depnevrsTag(h, td, hgflags, RPMTAG_CONFLICTNAME);
+}
+
+static int filenlinksTag(Header h, rpmtd td, headerGetFlags hgflags)
+{
+    rpmfi fi = rpmfiNew(NULL, h, RPMTAG_BASENAMES, RPMFI_NOHEADER);
+    rpm_count_t fc = rpmfiFC(fi);
+
+    if (fc > 0) {
+	uint32_t *nlinks = xmalloc(fc * sizeof(*nlinks));
+	int ix;
+	while ((ix = rpmfiNext(fi)) >= 0) {
+	    nlinks[ix] = rpmfiFNlink(fi);
+	}
+	td->data = nlinks;
+	td->type = RPM_INT32_TYPE;
+	td->count = fc;
+	td->flags = RPMTD_ALLOCED;
+    }
+
+    rpmfiFree(fi);
+    return (fc > 0);
+}
+
 static const struct headerTagFunc_s rpmHeaderTagExtensions[] = {
     { RPMTAG_GROUP,		groupTag },
     { RPMTAG_DESCRIPTION,	descriptionTag },
@@ -686,6 +873,16 @@ static const struct headerTagFunc_s rpmHeaderTagExtensions[] = {
     { RPMTAG_HEADERCOLOR,	headercolorTag },
     { RPMTAG_VERBOSE,		verboseTag },
     { RPMTAG_EPOCHNUM,		epochnumTag },
+    { RPMTAG_INSTFILENAMES,	instfilenamesTag },
+    { RPMTAG_REQUIRENEVRS,	requirenevrsTag },
+    { RPMTAG_RECOMMENDNEVRS,	recommendnevrsTag},
+    { RPMTAG_SUGGESTNEVRS,	suggestnevrsTag},
+    { RPMTAG_SUPPLEMENTNEVRS,	supplementnevrsTag},
+    { RPMTAG_ENHANCENEVRS,	enhancenevrsTag},
+    { RPMTAG_PROVIDENEVRS,	providenevrsTag },
+    { RPMTAG_OBSOLETENEVRS,	obsoletenevrsTag },
+    { RPMTAG_CONFLICTNEVRS,	conflictnevrsTag },
+    { RPMTAG_FILENLINKS,	filenlinksTag },
     { 0, 			NULL }
 };
 
