@@ -1,23 +1,13 @@
-/** \ingroup py_c
- * \file python/rpmmacro-py.c
- */
-
-#include "system.h"
-
-#include <structmember.h>
+#include "rpmsystem-py.h"
 
 #include <rpm/rpmmacro.h>
 
 #include "rpmmacro-py.h"
 
-#include "debug.h"
-
-/**
- */
 PyObject *
 rpmmacro_AddMacro(PyObject * self, PyObject * args, PyObject * kwds)
 {
-    char * name, * val;
+    const char * name, * val;
     char * kwlist[] = {"name", "value", NULL};
 
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "ss:AddMacro", kwlist,
@@ -26,16 +16,13 @@ rpmmacro_AddMacro(PyObject * self, PyObject * args, PyObject * kwds)
 
     addMacro(NULL, name, NULL, val, -1);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
-/**
- */
 PyObject *
 rpmmacro_DelMacro(PyObject * self, PyObject * args, PyObject * kwds)
 {
-    char * name;
+    const char * name;
     char * kwlist[] = {"name", NULL};
 
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "s:DelMacro", kwlist, &name))
@@ -43,24 +30,27 @@ rpmmacro_DelMacro(PyObject * self, PyObject * args, PyObject * kwds)
 
     delMacro(NULL, name);
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 
-/**
- */
 PyObject * 
 rpmmacro_ExpandMacro(PyObject * self, PyObject * args, PyObject * kwds)
 {
-    char *macro, *str;
+    const char *macro;
     PyObject *res;
+    int num = 0;
+    char * kwlist[] = {"macro", "numeric", NULL};
 
-    if (!PyArg_ParseTuple(args, "s", &macro))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|i", kwlist, &macro, &num))
         return NULL;
 
-    str = rpmExpand(macro, NULL);
-    res = Py_BuildValue("s", str);
-    free(str);
+    if (num) {
+	res = Py_BuildValue("i", rpmExpandNumeric(macro));
+    } else {
+	char *str = rpmExpand(macro, NULL);
+	res = Py_BuildValue("s", str);
+	free(str);
+    }
     return res;
 }
 
