@@ -1,17 +1,10 @@
-/** \ingroup py_c
- * \file python/rpmmi-py.c
- */
+#include "rpmsystem-py.h"
 
-#include "system.h"
-
-#include <rpmlib.h>
-#include <rpmdb.h>
+#include <rpm/rpmdb.h>
 
 #include "rpmmi-py.h"
 #include "header-py.h"
 
-#include "debug.h"
-
 /** \ingroup python
  * \class Rpmmi
  * \brief A python rpm.mi match iterator object represents the result of a
@@ -61,23 +54,19 @@
  *
  */
 
-/**
+/** \ingroup python
+ * \name Class: Rpmmi
  */
-static PyObject *
-rpmmi_iter(rpmmiObject * s)
-	/*@*/
-{
-    Py_INCREF(s);
-    return (PyObject *)s;
-}
 
-/**
- */
-/*@null@*/
+struct rpmmiObject_s {
+    PyObject_HEAD
+    PyObject *md_dict;		/*!< to look like PyModuleObject */
+    PyObject *ref;		/* for db/ts refcounting */
+    rpmdbMatchIterator mi;
+} ;
+
 static PyObject *
 rpmmi_iternext(rpmmiObject * s)
-	/*@globals rpmGlobalMacroContext @*/
-	/*@modifies s, rpmGlobalMacroContext @*/
 {
     Header h;
 
@@ -85,39 +74,11 @@ rpmmi_iternext(rpmmiObject * s)
 	s->mi = rpmdbFreeIterator(s->mi);
 	return NULL;
     }
-    return (PyObject *) hdr_Wrap(h);
+    return hdr_Wrap(&hdr_Type, h);
 }
 
-/**
- */
-/*@null@*/
 static PyObject *
-rpmmi_Next(rpmmiObject * s)
-	/*@globals rpmGlobalMacroContext, _Py_NoneStruct @*/
-	/*@modifies s, rpmGlobalMacroContext, _Py_NoneStruct @*/
-{
-    PyObject * result;
-
-    result = rpmmi_iternext(s);
-
-    if (result == NULL) {
-        Py_INCREF(Py_None);
-        return Py_None;
-    }
-    return result;
-}
-
-/** \ingroup python
- * \name Class: Rpmmi
- */
-/*@{*/
-
-/**
- */
-/*@null@*/
-static PyObject *
-rpmmi_Instance(rpmmiObject * s)
-	/*@*/
+rpmmi_Instance(rpmmiObject * s, PyObject * unused)
 {
     int rc = 0;
 
@@ -127,108 +88,84 @@ rpmmi_Instance(rpmmiObject * s)
     return Py_BuildValue("i", rc);
 }
 
-/**
- */
-/*@null@*/
 static PyObject *
-rpmmi_Count(rpmmiObject * s)
-	/*@*/
+rpmmi_Count(rpmmiObject * s, PyObject * unused)
 {
-    int rc = 0;
-
-    if (s->mi != NULL)
-	rc = rpmdbGetIteratorCount(s->mi);
-
-    return Py_BuildValue("i", rc);
+    DEPRECATED_METHOD("use len(mi) instead");
+    return Py_BuildValue("i", PyMapping_Size((PyObject *)s));
 }
 
-/**
- */
-/*@null@*/
 static PyObject *
 rpmmi_Pattern(rpmmiObject * s, PyObject * args, PyObject * kwds)
-	/*@globals rpmGlobalMacroContext, _Py_NoneStruct @*/
-	/*@modifies s, rpmGlobalMacroContext, _Py_NoneStruct @*/
 {
-    PyObject *TagN = NULL;
     int type;
-    char * pattern;
-    rpmTag tag;
+    const char * pattern;
+    rpmTagVal tag;
     char * kwlist[] = {"tag", "type", "patern", NULL};
 
-    if (!PyArg_ParseTupleAndKeywords(args, kwds, "Ois:Pattern", kwlist,
-	    &TagN, &type, &pattern))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "O&is:Pattern", kwlist,
+	    tagNumFromPyObject, &tag, &type, &pattern))
 	return NULL;
 
-    if ((tag = tagNumFromPyObject (TagN)) == -1) {
-	PyErr_SetString(PyExc_TypeError, "unknown tag type");
-	return NULL;
-    }
-
     rpmdbSetIteratorRE(s->mi, tag, type, pattern);
 
-    Py_INCREF (Py_None);
-    return Py_None;
-
+    Py_RETURN_NONE;
 }
 
-/*@}*/
-
-/** \ingroup py_c
- */
-/*@-fullinitblock@*/
-/*@unchecked@*/ /*@observer@*/
 static struct PyMethodDef rpmmi_methods[] = {
-    {"next",	    (PyCFunction) rpmmi_Next,		METH_NOARGS,
-"mi.next() -> hdr\n\
-- Retrieve next header that matches. Iterate directly in python if possible.\n" },
     {"instance",    (PyCFunction) rpmmi_Instance,	METH_NOARGS,
 	NULL },
     {"count",       (PyCFunction) rpmmi_Count,		METH_NOARGS,
-	NULL },
+"Deprecated, use len(mi) instead.\n" },
     {"pattern",	    (PyCFunction) rpmmi_Pattern,	METH_VARARGS|METH_KEYWORDS,
 "mi.pattern(TagN, mire_type, pattern)\n\
 - Set a secondary match pattern on tags from retrieved header.\n" },
     {NULL,		NULL}		/* sentinel */
 };
-/*@=fullinitblock@*/
 
-/** \ingroup py_c
- */
-static void rpmmi_dealloc(/*@only@*/ /*@null@*/ rpmmiObject * s)
-	/*@globals rpmGlobalMacroContext @*/
-	/*@modifies s, rpmGlobalMacroContext @*/
+static void rpmmi_dealloc(rpmmiObject * s)
 {
-    if (s) {
-	s->mi = rpmdbFreeIterator(s->mi);
-	PyObject_Del(s);
-    }
+    s->mi = rpmdbFreeIterator(s->mi);
+    Py_DECREF(s->ref);
+    Py_TYPE(s)->tp_free((PyObject *)s);
 }
 
-static PyObject * rpmmi_getattro(PyObject * o, PyObject * n)
-	/*@*/
+static Py_ssize_t rpmmi_length(rpmmiObject * s)
 {
-    return PyObject_GenericGetAttr(o, n);
+    return s->mi ? rpmdbGetIteratorCount(s->mi) : 0;
 }
 
-static int rpmmi_setattro(PyObject * o, PyObject * n, PyObject * v)
-	/*@*/
+static int rpmmi_bool(rpmmiObject *s)
 {
-    return PyObject_GenericSetAttr(o, n, v);
+    return (s->mi != NULL);
 }
 
-/**
- */
-/*@unchecked@*/ /*@observer@*/
+PyMappingMethods rpmmi_as_mapping = {
+    (lenfunc) rpmmi_length,		/* mp_length */
+    0,
+};
+
+static PyNumberMethods rpmmi_as_number = {
+	0, /* nb_add */
+	0, /* nb_subtract */
+	0, /* nb_multiply */
+#if PY_MAJOR_VERSION < 3
+	0, /* nb_divide */
+#endif
+	0, /* nb_remainder */
+	0, /* nb_divmod */
+	0, /* nb_power */
+	0, /* nb_negative */
+	0, /* nb_positive */
+	0, /* nb_absolute */
+	(inquiry)rpmmi_bool, /* nb_bool/nonzero */
+};
+
 static char rpmmi_doc[] =
 "";
 
-/** \ingroup py_c
- */
-/*@-fullinitblock@*/
 PyTypeObject rpmmi_Type = {
-	PyObject_HEAD_INIT(&PyType_Type)
-	0,				/* ob_size */
+	PyVarObject_HEAD_INIT(&PyType_Type, 0)
 	"rpm.mi",			/* tp_name */
 	sizeof(rpmmiObject),		/* tp_size */
 	0,				/* tp_itemsize */
@@ -238,23 +175,22 @@ PyTypeObject rpmmi_Type = {
 	0,				/* tp_setattr */
 	0,				/* tp_compare */
 	0,				/* tp_repr */
-	0,				/* tp_as_number */
+	&rpmmi_as_number,		/* tp_as_number */
 	0,				/* tp_as_sequence */
-	0,				/* tp_as_mapping */
+	&rpmmi_as_mapping,		/* tp_as_mapping */
 	0,				/* tp_hash */
 	0,				/* tp_call */
 	0,				/* tp_str */
-	(getattrofunc) rpmmi_getattro,	/* tp_getattro */
-	(setattrofunc) rpmmi_setattro,	/* tp_setattro */
+	PyObject_GenericGetAttr,	/* tp_getattro */
+	PyObject_GenericSetAttr,	/* tp_setattro */
 	0,				/* tp_as_buffer */
-	Py_TPFLAGS_DEFAULT,		/* tp_flags */
+	Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE,	/* tp_flags */
 	rpmmi_doc,			/* tp_doc */
-#if Py_TPFLAGS_HAVE_ITER
 	0,				/* tp_traverse */
 	0,				/* tp_clear */
 	0,				/* tp_richcompare */
 	0,				/* tp_weaklistoffset */
-	(getiterfunc) rpmmi_iter,	/* tp_iter */
+	PyObject_SelfIter,		/* tp_iter */
 	(iternextfunc) rpmmi_iternext,	/* tp_iternext */
 	rpmmi_methods,			/* tp_methods */
 	0,				/* tp_members */
@@ -269,18 +205,16 @@ PyTypeObject rpmmi_Type = {
 	0,				/* tp_new */
 	0,				/* tp_free */
 	0,				/* tp_is_gc */
-#endif
 };
-/*@=fullinitblock@*/
 
-rpmmiObject * rpmmi_Wrap(rpmdbMatchIterator mi)
+PyObject * rpmmi_Wrap(PyTypeObject *subtype, rpmdbMatchIterator mi, PyObject *s)
 {
-    rpmmiObject * mio = (rpmmiObject *) PyObject_New(rpmmiObject, &rpmmi_Type);
+    rpmmiObject * mio = (rpmmiObject *)subtype->tp_alloc(subtype, 0);
+    if (mio == NULL) return NULL;
 
-    if (mio == NULL) {
-        PyErr_SetString(pyrpmError, "out of memory creating rpmmiObject");
-        return NULL;
-    }
     mio->mi = mi;
-    return mio;
+    mio->ref = s;
+    Py_INCREF(mio->ref);
+    return (PyObject *) mio;
 }
+
