@@ -5,21 +5,24 @@
 #include <rpm/rpmdb.h>
 #include <rpm/rpmsq.h>
 #include <rpm/rpmlog.h>
+#include <rpm/rpmmacro.h>
 
 #include "header-py.h"
+#include "rpmarchive-py.h"
 #include "rpmds-py.h"
 #include "rpmfd-py.h"
 #include "rpmfi-py.h"
+#include "rpmfiles-py.h"
 #include "rpmkeyring-py.h"
 #include "rpmmi-py.h"
+#include "rpmii-py.h"
 #include "rpmps-py.h"
 #include "rpmmacro-py.h"
+#include "rpmstrpool-py.h"
 #include "rpmtd-py.h"
 #include "rpmte-py.h"
 #include "rpmts-py.h"
 
-#include "debug.h"
-
 /** \ingroup python
  * \name Module: rpm
  */
@@ -28,7 +31,7 @@ PyObject * pyrpmError;
 
 static PyObject * archScore(PyObject * self, PyObject * arg)
 {
-    char * arch;
+    const char * arch;
 
     if (!PyArg_Parse(arg, "s", &arch))
 	return NULL;
@@ -115,6 +118,22 @@ static PyObject * doLog(PyObject * self, PyObject * args, PyObject *kwds)
     Py_RETURN_NONE;
 }
 
+static PyObject * reloadConfig(PyObject * self, PyObject * args, PyObject *kwds)
+{
+    const char * target = NULL;
+    char * kwlist[] = { "target", NULL };
+    int rc;
+    
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "|s", kwlist, &target))
+        return NULL;
+
+    rpmFreeMacros(NULL);
+    rpmFreeRpmrc();
+    rc = rpmReadConfigFiles(NULL, target) ;
+
+    return PyBool_FromLong(rc == 0);
+}
+
 static PyMethodDef rpmModuleMethods[] = {
     { "addMacro", (PyCFunction) rpmmacro_AddMacro, METH_VARARGS|METH_KEYWORDS,
 	NULL },
@@ -149,6 +168,8 @@ static PyMethodDef rpmModuleMethods[] = {
 	NULL },
     { "setStats", (PyCFunction) setStats, METH_O,
 	NULL },
+    { "reloadConfig", (PyCFunction) reloadConfig, METH_VARARGS|METH_KEYWORDS,
+	NULL },
 
     { NULL }
 } ;
@@ -173,7 +194,7 @@ static void addRpmTags(PyObject *module)
     rpmtd names = rpmtdNew();
     rpmTagGetNames(names, 1);
     const char *tagname, *shortname;
-    rpmTag tagval;
+    rpmTagVal tagval;
 
     while ((tagname = rpmtdNextString(names))) {
 	shortname = tagname + strlen("RPMTAG_");
@@ -197,13 +218,18 @@ static void addRpmTags(PyObject *module)
 static int prepareInitModule(void)
 {
     if (PyType_Ready(&hdr_Type) < 0) return 0;
+    if (PyType_Ready(&rpmarchive_Type) < 0) return 0;
     if (PyType_Ready(&rpmds_Type) < 0) return 0;
     if (PyType_Ready(&rpmfd_Type) < 0) return 0;
     if (PyType_Ready(&rpmfi_Type) < 0) return 0;
+    if (PyType_Ready(&rpmfile_Type) < 0) return 0;
+    if (PyType_Ready(&rpmfiles_Type) < 0) return 0;
     if (PyType_Ready(&rpmKeyring_Type) < 0) return 0;
     if (PyType_Ready(&rpmmi_Type) < 0) return 0;
+    if (PyType_Ready(&rpmii_Type) < 0) return 0;
     if (PyType_Ready(&rpmProblem_Type) < 0) return 0;
     if (PyType_Ready(&rpmPubkey_Type) < 0) return 0;
+    if (PyType_Ready(&rpmstrPool_Type) < 0) return 0;
 #if 0
     if (PyType_Ready(&rpmtd_Type) < 0) return 0;
 #endif
@@ -275,7 +301,9 @@ static int initModule(PyObject *m)
     if (Py_AtExit(rpm_exithook) == -1)
         return 0;
 
-    rpmReadConfigFiles(NULL, NULL);
+    /* failure to initialize rpm (crypto and all) is rather fatal too... */
+    if (rpmReadConfigFiles(NULL, NULL) == -1)
+	return 0;
 
     d = PyModule_GetDict(m);
 
@@ -286,6 +314,9 @@ static int initModule(PyObject *m)
     Py_INCREF(&hdr_Type);
     PyModule_AddObject(m, "hdr", (PyObject *) &hdr_Type);
 
+    Py_INCREF(&rpmarchive_Type);
+    PyModule_AddObject(m, "archive", (PyObject *) &rpmarchive_Type);
+
     Py_INCREF(&rpmds_Type);
     PyModule_AddObject(m, "ds", (PyObject *) &rpmds_Type);
 
@@ -295,18 +326,30 @@ static int initModule(PyObject *m)
     Py_INCREF(&rpmfi_Type);
     PyModule_AddObject(m, "fi", (PyObject *) &rpmfi_Type);
 
+    Py_INCREF(&rpmfile_Type);
+    PyModule_AddObject(m, "file", (PyObject *) &rpmfile_Type);
+
+    Py_INCREF(&rpmfiles_Type);
+    PyModule_AddObject(m, "files", (PyObject *) &rpmfiles_Type);
+
     Py_INCREF(&rpmKeyring_Type);
     PyModule_AddObject(m, "keyring", (PyObject *) &rpmKeyring_Type);
 
     Py_INCREF(&rpmmi_Type);
     PyModule_AddObject(m, "mi", (PyObject *) &rpmmi_Type);
 
+    Py_INCREF(&rpmii_Type);
+    PyModule_AddObject(m, "ii", (PyObject *) &rpmii_Type);
+
     Py_INCREF(&rpmProblem_Type);
     PyModule_AddObject(m, "prob", (PyObject *) &rpmProblem_Type);
 
     Py_INCREF(&rpmPubkey_Type);
     PyModule_AddObject(m, "pubkey", (PyObject *) &rpmPubkey_Type);
 
+    Py_INCREF(&rpmstrPool_Type);
+    PyModule_AddObject(m, "strpool", (PyObject *) &rpmstrPool_Type);
+
 #if 0
     Py_INCREF(&rpmtd_Type);
     PyModule_AddObject(m, "td", (PyObject *) &rpmtd_Type);
@@ -322,6 +365,9 @@ static int initModule(PyObject *m)
 
     PyModule_AddStringConstant(m, "__version__", RPMVERSION);
 
+    PyModule_AddObject(m, "header_magic",
+		PyBytes_FromStringAndSize((const char *)rpm_header_magic, 8));
+
 #define REGISTER_ENUM(val) PyModule_AddIntConstant(m, #val, val)
 
     REGISTER_ENUM(RPMTAG_NOT_FOUND);
@@ -340,13 +386,13 @@ static int initModule(PyObject *m)
 
     REGISTER_ENUM(RPMFILE_CONFIG);
     REGISTER_ENUM(RPMFILE_DOC);
+    REGISTER_ENUM(RPMFILE_ICON);
     REGISTER_ENUM(RPMFILE_MISSINGOK);
     REGISTER_ENUM(RPMFILE_NOREPLACE);
+    REGISTER_ENUM(RPMFILE_SPECFILE);
     REGISTER_ENUM(RPMFILE_GHOST);
     REGISTER_ENUM(RPMFILE_LICENSE);
     REGISTER_ENUM(RPMFILE_README);
-    REGISTER_ENUM(RPMFILE_EXCLUDE);
-    REGISTER_ENUM(RPMFILE_UNPATCHED);
     REGISTER_ENUM(RPMFILE_PUBKEY);
 
     REGISTER_ENUM(RPMDEP_SENSE_REQUIRES);
@@ -356,7 +402,9 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMSENSE_LESS);
     REGISTER_ENUM(RPMSENSE_GREATER);
     REGISTER_ENUM(RPMSENSE_EQUAL);
+    REGISTER_ENUM(RPMSENSE_POSTTRANS);
     REGISTER_ENUM(RPMSENSE_PREREQ);
+    REGISTER_ENUM(RPMSENSE_PRETRANS);
     REGISTER_ENUM(RPMSENSE_INTERP);
     REGISTER_ENUM(RPMSENSE_SCRIPT_PRE);
     REGISTER_ENUM(RPMSENSE_SCRIPT_POST);
@@ -368,14 +416,9 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMSENSE_TRIGGERIN);
     REGISTER_ENUM(RPMSENSE_TRIGGERUN);
     REGISTER_ENUM(RPMSENSE_TRIGGERPOSTUN);
-    REGISTER_ENUM(RPMSENSE_SCRIPT_PREP);
-    REGISTER_ENUM(RPMSENSE_SCRIPT_BUILD);
-    REGISTER_ENUM(RPMSENSE_SCRIPT_INSTALL);
-    REGISTER_ENUM(RPMSENSE_SCRIPT_CLEAN);
     REGISTER_ENUM(RPMSENSE_RPMLIB);
     REGISTER_ENUM(RPMSENSE_TRIGGERPREIN);
     REGISTER_ENUM(RPMSENSE_KEYRING);
-    REGISTER_ENUM(RPMSENSE_PATCHES);
     REGISTER_ENUM(RPMSENSE_CONFIG);
 
     REGISTER_ENUM(RPMTRANS_FLAG_TEST);
@@ -385,6 +428,7 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERS);
     REGISTER_ENUM(RPMTRANS_FLAG_NODOCS);
     REGISTER_ENUM(RPMTRANS_FLAG_ALLFILES);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOPLUGINS);
     REGISTER_ENUM(RPMTRANS_FLAG_KEEPOBSOLETE);
     REGISTER_ENUM(RPMTRANS_FLAG_NOCONTEXTS);
     REGISTER_ENUM(RPMTRANS_FLAG_REPACKAGE);
@@ -397,6 +441,8 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMTRANS_FLAG_NOPREUN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOPOSTUN);
     REGISTER_ENUM(RPMTRANS_FLAG_NOTRIGGERPOSTUN);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOPRETRANS);
+    REGISTER_ENUM(RPMTRANS_FLAG_NOPOSTTRANS);
     REGISTER_ENUM(RPMTRANS_FLAG_NOMD5);
     REGISTER_ENUM(RPMTRANS_FLAG_NOFILEDIGEST);
     REGISTER_ENUM(RPMTRANS_FLAG_NOSUGGEST);
@@ -430,6 +476,9 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMCALLBACK_UNPACK_ERROR);
     REGISTER_ENUM(RPMCALLBACK_CPIO_ERROR);
     REGISTER_ENUM(RPMCALLBACK_SCRIPT_ERROR);
+    REGISTER_ENUM(RPMCALLBACK_SCRIPT_START);
+    REGISTER_ENUM(RPMCALLBACK_SCRIPT_STOP);
+    REGISTER_ENUM(RPMCALLBACK_INST_STOP);
 
     REGISTER_ENUM(RPMPROB_BADARCH);
     REGISTER_ENUM(RPMPROB_BADOS);
@@ -442,6 +491,7 @@ static int initModule(PyObject *m)
     REGISTER_ENUM(RPMPROB_OLDPACKAGE);
     REGISTER_ENUM(RPMPROB_DISKSPACE);
     REGISTER_ENUM(RPMPROB_DISKNODES);
+    REGISTER_ENUM(RPMPROB_OBSOLETES);
 
     REGISTER_ENUM(VERIFY_DIGEST);
     REGISTER_ENUM(VERIFY_SIGNATURE);
@@ -481,6 +531,19 @@ static int initModule(PyObject *m)
 
     REGISTER_ENUM(RPMDBI_PACKAGES);
     REGISTER_ENUM(RPMDBI_LABEL);
+    REGISTER_ENUM(RPMDBI_INSTFILENAMES);
+    REGISTER_ENUM(RPMDBI_NAME);
+    REGISTER_ENUM(RPMDBI_BASENAMES);
+    REGISTER_ENUM(RPMDBI_GROUP);
+    REGISTER_ENUM(RPMDBI_REQUIRENAME);
+    REGISTER_ENUM(RPMDBI_PROVIDENAME);
+    REGISTER_ENUM(RPMDBI_CONFLICTNAME);
+    REGISTER_ENUM(RPMDBI_OBSOLETENAME);
+    REGISTER_ENUM(RPMDBI_TRIGGERNAME);
+    REGISTER_ENUM(RPMDBI_DIRNAMES);
+    REGISTER_ENUM(RPMDBI_INSTALLTID);
+    REGISTER_ENUM(RPMDBI_SIGMD5);
+    REGISTER_ENUM(RPMDBI_SHA1HEADER);
 
     REGISTER_ENUM(HEADERCONV_EXPANDFILELIST);
     REGISTER_ENUM(HEADERCONV_COMPRESSFILELIST);
