@@ -3,12 +3,10 @@
 #include "header-py.h"	/* XXX tagNumFromPyObject */
 #include "rpmds-py.h"
 #include "rpmfi-py.h"
+#include "rpmfiles-py.h"
 #include "rpmte-py.h"
 #include "rpmps-py.h"
 
-#include "debug.h"
-
-
 /** \ingroup python
  * \name Class: Rpmte
  * \class Rpmte
@@ -38,8 +36,9 @@
  * - te.Key()	Return the associated opaque key, i.e. 2nd arg ts.addInstall().
  * - te.DS(tag)	Return package dependency set.
  * @param tag	'Providename', 'Requirename', 'Obsoletename', 'Conflictname'
- * - te.FI(tag)	Return package file info set.
+ * - te.FI(tag)	Return package file info set iterator (deprecated).
  * @param tag	'Basenames'
+ * - te.Files()	Return package file info set.
  */
 
 struct rpmteObject_s {
@@ -49,83 +48,87 @@ struct rpmteObject_s {
 };
 
 static PyObject *
-rpmte_TEType(rpmteObject * s)
+rpmte_TEType(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("i", rpmteType(s->te));
 }
 
 static PyObject *
-rpmte_N(rpmteObject * s)
+rpmte_N(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteN(s->te));
 }
 
 static PyObject *
-rpmte_E(rpmteObject * s)
+rpmte_E(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteE(s->te));
 }
 
 static PyObject *
-rpmte_V(rpmteObject * s)
+rpmte_V(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteV(s->te));
 }
 
 static PyObject *
-rpmte_R(rpmteObject * s)
+rpmte_R(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteR(s->te));
 }
 
 static PyObject *
-rpmte_A(rpmteObject * s)
+rpmte_A(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteA(s->te));
 }
 
 static PyObject *
-rpmte_O(rpmteObject * s)
+rpmte_O(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteO(s->te));
 }
 
 static PyObject *
-rpmte_NEVR(rpmteObject * s)
+rpmte_NEVR(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteNEVR(s->te));
 }
 
 static PyObject *
-rpmte_NEVRA(rpmteObject * s)
+rpmte_NEVRA(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("s", rpmteNEVRA(s->te));
 }
 
 static PyObject *
-rpmte_Color(rpmteObject * s)
+rpmte_Color(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("i", rpmteColor(s->te));
 }
 
 static PyObject *
-rpmte_PkgFileSize(rpmteObject * s)
+rpmte_PkgFileSize(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("L", rpmtePkgFileSize(s->te));
 }
 
 static PyObject *
-rpmte_Parent(rpmteObject * s)
+rpmte_Parent(rpmteObject * s, PyObject * unused)
 {
-    return Py_BuildValue("i", rpmteParent(s->te));
+    rpmte parent = rpmteParent(s->te);
+    if (parent)
+	return rpmte_Wrap(&rpmte_Type, parent);
+
+    Py_RETURN_NONE;
 }
 
-static PyObject * rpmte_Failed(rpmteObject * s)
+static PyObject * rpmte_Failed(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("i", rpmteFailed(s->te));
 }
 
-static PyObject * rpmte_Problems(rpmteObject * s)
+static PyObject * rpmte_Problems(rpmteObject * s, PyObject * unused)
 {
     rpmps ps = rpmteProblems(s->te);
     PyObject *problems = rpmps_AsList(ps);
@@ -135,28 +138,26 @@ static PyObject * rpmte_Problems(rpmteObject * s)
 
 /*
 static PyObject *
-rpmte_DependsOnKey(rpmteObject * s)
+rpmte_DependsOnKey(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("i", rpmteDependsOnKey(s->te));
 }
 */
 
 static PyObject *
-rpmte_DBOffset(rpmteObject * s)
+rpmte_DBOffset(rpmteObject * s, PyObject * unused)
 {
     return Py_BuildValue("i", rpmteDBOffset(s->te));
 }
 
 static PyObject *
-rpmte_Key(rpmteObject * s)
+rpmte_Key(rpmteObject * s, PyObject * unused)
 {
-    PyObject * Key;
-
     /* XXX how to insure this is a PyObject??? */
-    Key = (PyObject *) rpmteKey(s->te);
+    PyObject * Key = (PyObject *) rpmteKey(s->te);
     if (Key == NULL)
-      Key = Py_None;
-	Py_INCREF(Key);
+	Key = Py_None;
+    Py_INCREF(Key);
     return Key;
 }
 
@@ -183,6 +184,8 @@ rpmte_FI(rpmteObject * s, PyObject * args, PyObject * kwds)
 {
     rpmfi fi;
 
+    DEPRECATED_METHOD("use .Files() instead");
+
     fi = rpmteFI(s->te);
     if (fi == NULL) {
 	Py_RETURN_NONE;
@@ -190,6 +193,15 @@ rpmte_FI(rpmteObject * s, PyObject * args, PyObject * kwds)
     return rpmfi_Wrap(&rpmfi_Type, rpmfiLink(fi));
 }
 
+static PyObject *
+rpmte_Files(rpmteObject * s, PyObject * args, PyObject * kwds)
+{
+    rpmfiles files = rpmteFiles(s->te);
+    if (files == NULL) {
+	Py_RETURN_NONE;
+    }
+    return rpmfiles_Wrap(&rpmfiles_Type, files);
+}
 static struct PyMethodDef rpmte_methods[] = {
     {"Type",	(PyCFunction)rpmte_TEType,	METH_NOARGS,
 "te.Type() -> Type\n\
@@ -240,7 +252,10 @@ static struct PyMethodDef rpmte_methods[] = {
 	'Providename', 'Requirename', 'Obsoletename', 'Conflictname'\n" },
     {"FI",	(PyCFunction)rpmte_FI,		METH_VARARGS|METH_KEYWORDS,
 "te.FI(TagN) -> FI\n\
-- Return the TagN dependency set (or None). TagN must be 'Basenames'.\n" },
+- Return file info iterator of element. Deprecated, use .Files() instead.\n" },
+    {"Files",	(PyCFunction)rpmte_Files,	METH_NOARGS,
+"te.Files() -> files\n\
+- Return file info set of element.\n" },
     {NULL,		NULL}		/* sentinel */
 };
 
