@@ -4,6 +4,8 @@
 
 #include "system.h"
 #include <stdarg.h>
+#include <pthread.h>
+#include <errno.h>
 #ifdef HAVE_GETOPT_H
 #include <getopt.h>
 #else
@@ -34,21 +36,29 @@ extern int optind;
 
 #include "debug.h"
 
+enum macroFlags_e {
+    ME_NONE	= 0,
+    ME_AUTO	= (1 << 0),
+    ME_USED	= (1 << 1),
+};
+
 /*! The structure used to store a macro. */
 struct rpmMacroEntry_s {
     struct rpmMacroEntry_s *prev;/*!< Macro entry stack. */
-    char *name;   	/*!< Macro name. */
-    char *opts;   	/*!< Macro parameters (a la getopt) */
-    char *body;   	/*!< Macro body. */
-    int used;           /*!< No. of expansions. */
+    const char *name;  	/*!< Macro name. */
+    const char *opts;  	/*!< Macro parameters (a la getopt) */
+    const char *body;	/*!< Macro body. */
+    int flags;		/*!< Macro state bits. */
     int level;          /*!< Scoping level. */
+    char arena[];   	/*!< String arena. */
 };
 
 /*! The structure used to store the set of macros in a context. */
 struct rpmMacroContext_s {
-    rpmMacroEntry *macroTable;  /*!< Macro entry table for context. */
-    int macrosAllocated;/*!< No. of allocated macros. */
-    int firstFree;      /*!< No. of macros. */
+    rpmMacroEntry *tab;  /*!< Macro entry table (array of pointers). */
+    int n;      /*!< No. of macros. */
+    pthread_mutex_t lock;
+    pthread_mutexattr_t lockattr;
 };
 
 
@@ -58,6 +68,29 @@ rpmMacroContext rpmGlobalMacroContext = &rpmGlobalMacroContext_s;
 static struct rpmMacroContext_s rpmCLIMacroContext_s;
 rpmMacroContext rpmCLIMacroContext = &rpmCLIMacroContext_s;
 
+/*
+ * The macro engine internals do not require recursive mutexes but Lua
+ * macro bindings which can get called from the internals use the external
+ * interfaces which do perform locking. Until that is fixed somehow
+ * we'll just have to settle for recursive mutexes.
+ * Unfortunately POSIX doesn't specify static initializers for recursive
+ * mutexes so we need to have a separate PTHREAD_ONCE initializer just
+ * to initialize the otherwise static macro context mutexes. Pooh.
+ */
+static pthread_once_t locksInitialized = PTHREAD_ONCE_INIT;
+
+static void initLocks(void)
+{
+    rpmMacroContext mcs[] = { rpmGlobalMacroContext, rpmCLIMacroContext, NULL };
+
+    for (rpmMacroContext *mcp = mcs; *mcp; mcp++) {
+	rpmMacroContext mc = *mcp;
+	pthread_mutexattr_init(&mc->lockattr);
+	pthread_mutexattr_settype(&mc->lockattr, PTHREAD_MUTEX_RECURSIVE);
+	pthread_mutex_init(&mc->lock, &mc->lockattr);
+    }
+}
+
 /**
  * Macro expansion state.
  */
@@ -80,110 +113,28 @@ static int print_macro_trace = _PRINT_MACRO_TRACE;
 #define	_PRINT_EXPAND_TRACE	0
 static int print_expand_trace = _PRINT_EXPAND_TRACE;
 
-#define	MACRO_CHUNK_SIZE	16
-
 /* forward ref */
 static int expandMacro(MacroBuf mb, const char *src, size_t slen);
+static void pushMacro(rpmMacroContext mc,
+	const char * n, const char * o, const char * b, int level, int flags);
+static void popMacro(rpmMacroContext mc, const char * n);
+static int loadMacroFile(rpmMacroContext mc, const char * fn);
 
 /* =============================================================== */
 
-/**
- * Compare macro entries by name (qsort/bsearch).
- * @param ap		1st macro entry
- * @param bp		2nd macro entry
- * @return		result of comparison
- */
-static int
-compareMacroName(const void * ap, const void * bp)
+static rpmMacroContext rpmmctxAcquire(rpmMacroContext mc)
 {
-    rpmMacroEntry ame = *((const rpmMacroEntry *)ap);
-    rpmMacroEntry bme = *((const rpmMacroEntry *)bp);
-
-    if (ame == NULL && bme == NULL)
-	return 0;
-    if (ame == NULL)
-	return 1;
-    if (bme == NULL)
-	return -1;
-    return strcmp(ame->name, bme->name);
+    if (mc == NULL)
+	mc = rpmGlobalMacroContext;
+    pthread_once(&locksInitialized, initLocks);
+    pthread_mutex_lock(&mc->lock);
+    return mc;
 }
 
-/**
- * Enlarge macro table.
- * @param mc		macro context
- */
-static void
-expandMacroTable(rpmMacroContext mc)
+static rpmMacroContext rpmmctxRelease(rpmMacroContext mc)
 {
-    if (mc->macroTable == NULL) {
-	mc->macrosAllocated = MACRO_CHUNK_SIZE;
-	mc->macroTable = (rpmMacroEntry *)
-	    xmalloc(sizeof(*(mc->macroTable)) * mc->macrosAllocated);
-	mc->firstFree = 0;
-    } else {
-	mc->macrosAllocated += MACRO_CHUNK_SIZE;
-	mc->macroTable = (rpmMacroEntry *)
-	    xrealloc(mc->macroTable, sizeof(*(mc->macroTable)) *
-			mc->macrosAllocated);
-    }
-    memset(&mc->macroTable[mc->firstFree], 0, MACRO_CHUNK_SIZE * sizeof(*(mc->macroTable)));
-}
-
-/**
- * Sort entries in macro table.
- * @param mc		macro context
- */
-static void
-sortMacroTable(rpmMacroContext mc)
-{
-    int i;
-
-    if (mc == NULL || mc->macroTable == NULL)
-	return;
-
-    qsort(mc->macroTable, mc->firstFree, sizeof(*(mc->macroTable)),
-		compareMacroName);
-
-    /* Empty pointers are now at end of table. Reset first free index. */
-    for (i = 0; i < mc->firstFree; i++) {
-	if (mc->macroTable[i] != NULL)
-	    continue;
-	mc->firstFree = i;
-	break;
-    }
-}
-
-void
-rpmDumpMacroTable(rpmMacroContext mc, FILE * fp)
-{
-    int nempty = 0;
-    int nactive = 0;
-
-    if (mc == NULL) mc = rpmGlobalMacroContext;
-    if (fp == NULL) fp = stderr;
-    
-    fprintf(fp, "========================\n");
-    if (mc->macroTable != NULL) {
-	int i;
-	for (i = 0; i < mc->firstFree; i++) {
-	    rpmMacroEntry me;
-	    if ((me = mc->macroTable[i]) == NULL) {
-		/* XXX this should never happen */
-		nempty++;
-		continue;
-	    }
-	    fprintf(fp, "%3d%c %s", me->level,
-			(me->used > 0 ? '=' : ':'), me->name);
-	    if (me->opts && *me->opts)
-		    fprintf(fp, "(%s)", me->opts);
-	    if (me->body && *me->body)
-		    fprintf(fp, "\t%s", me->body);
-	    fprintf(fp, "\n");
-	    nactive++;
-	}
-    }
-    fprintf(fp, _("======================== active %d empty %d\n"),
-		nactive, nempty);
+    pthread_mutex_unlock(&mc->lock);
+    return NULL;
 }
 
 /**
@@ -191,33 +142,41 @@ rpmDumpMacroTable(rpmMacroContext mc, FILE * fp)
  * @param mc		macro context
  * @param name		macro name
  * @param namelen	no. of bytes
+ * @param pos		found/insert position
  * @return		address of slot in macro table with name (or NULL)
  */
 static rpmMacroEntry *
-findEntry(rpmMacroContext mc, const char * name, size_t namelen)
+findEntry(rpmMacroContext mc, const char *name, size_t namelen, size_t *pos)
 {
-    rpmMacroEntry key, *ret;
-    struct rpmMacroEntry_s keybuf;
-    char namebuf[namelen+1];
-    const char *mname = name;
-
-    if (mc == NULL) mc = rpmGlobalMacroContext;
-    if (mc->macroTable == NULL || mc->firstFree == 0)
-	return NULL;
-
-    if (namelen > 0) {
-	strncpy(namebuf, name, namelen);
-	namebuf[namelen] = '\0';
-	mname = namebuf;
+    /* bsearch */
+    int cmp = 1;
+    size_t l = 0;
+    size_t u = mc->n;
+    size_t i = 0;
+    while (l < u) {
+	i = (l + u) / 2;
+	rpmMacroEntry me = mc->tab[i];
+	if (namelen == 0)
+	    cmp = strcmp(me->name, name);
+	else {
+	    cmp = strncmp(me->name, name, namelen);
+	    /* longer me->name compares greater */
+	    if (cmp == 0)
+		cmp = (me->name[namelen] != '\0');
+	}
+	if (cmp < 0)
+	    l = i + 1;
+	else if (cmp > 0)
+	    u = i;
+	else
+	    break;
     }
-    
-    key = &keybuf;
-    memset(key, 0, sizeof(*key));
-    key->name = (char *)mname;
-    ret = (rpmMacroEntry *) bsearch(&key, mc->macroTable, mc->firstFree,
-			sizeof(*(mc->macroTable)), compareMacroName);
-    /* XXX TODO: find 1st empty slot and return that */
-    return ret;
+
+    if (pos)
+	*pos = (cmp < 0) ? i + 1 : i;
+    if (cmp == 0)
+	return &mc->tab[i];
+    return NULL;
 }
 
 /* =============================================================== */
@@ -438,10 +397,11 @@ expandThis(MacroBuf mb, const char * src, size_t slen, char **target)
 static void mbAppend(MacroBuf mb, char c)
 {
     if (mb->nb < 1) {
-	mb->buf = xrealloc(mb->buf, mb->tpos + mb->nb + MACROBUFSIZ);
+	mb->buf = xrealloc(mb->buf, mb->tpos + MACROBUFSIZ + 1);
 	mb->nb += MACROBUFSIZ;
     }
     mb->buf[mb->tpos++] = c;
+    mb->buf[mb->tpos] = '\0';
     mb->nb--;
 }
 
@@ -449,10 +409,10 @@ static void mbAppendStr(MacroBuf mb, const char *str)
 {
     size_t len = strlen(str);
     if (len > mb->nb) {
-	mb->buf = xrealloc(mb->buf, mb->tpos + mb->nb + MACROBUFSIZ + len);
+	mb->buf = xrealloc(mb->buf, mb->tpos + mb->nb + MACROBUFSIZ + len + 1);
 	mb->nb += MACROBUFSIZ + len;
     }
-    memcpy(mb->buf+mb->tpos, str, len);
+    memcpy(mb->buf+mb->tpos, str, len + 1);
     mb->tpos += len;
     mb->nb -= len;
 }
@@ -479,14 +439,16 @@ doShellEscape(MacroBuf mb, const char * cmd, size_t clen)
 	rc = 1;
 	goto exit;
     }
+
+    size_t tpos = mb->tpos;
     while((c = fgetc(shf)) != EOF) {
 	mbAppend(mb, c);
     }
     (void) pclose(shf);
 
-    /* XXX delete trailing \r \n */
-    while (iseol(mb->buf[mb->tpos-1])) {
-	mb->buf[mb->tpos--] = '\0';
+    /* Delete trailing \r \n */
+    while (mb->tpos > tpos && iseol(mb->buf[mb->tpos-1])) {
+	mb->buf[--mb->tpos] = '\0';
 	mb->nb++;
     }
 
@@ -499,21 +461,22 @@ exit:
  * Parse (and execute) new macro definition.
  * @param mb		macro expansion state
  * @param se		macro definition to parse
+ * @param slen		length of se argument
  * @param level		macro recursion level
  * @param expandbody	should body be expanded?
  * @return		address to continue parsing
  */
 static const char *
-doDefine(MacroBuf mb, const char * se, int level, int expandbody)
+doDefine(MacroBuf mb, const char * se, size_t slen, int level, int expandbody)
 {
     const char *s = se;
-    size_t blen = MACROBUFSIZ;
-    char *buf = xmalloc(blen);
+    char *buf = xmalloc(slen + 3); /* Some leeway for termination issues... */
     char *n = buf, *ne = n;
     char *o = NULL, *oe;
     char *b, *be, *ebody = NULL;
     int c;
     int oc = ')';
+    const char *sbody; /* as-is body start */
 
     /* Copy name */
     COPYNAME(ne, s, c);
@@ -522,13 +485,20 @@ doDefine(MacroBuf mb, const char * se, int level, int expandbody)
     oe = ne + 1;
     if (*s == '(') {
 	s++;	/* skip ( */
-	o = oe;
-	COPYOPTS(oe, s, oc);
-	s++;	/* skip ) */
+	/* Options must be terminated with ')' */
+	if (strchr(s, ')')) {
+	    o = oe;
+	    COPYOPTS(oe, s, oc);
+	    s++;	/* skip ) */
+	} else {
+	    rpmlog(RPMLOG_ERR, _("Macro %%%s has unterminated opts\n"), n);
+	    goto exit;
+	}
     }
 
     /* Copy body, skipping over escaped newlines */
     b = be = oe + 1;
+    sbody = s;
     SKIPBLANK(s, c);
     if (c == '{') {	/* XXX permit silent {...} grouping */
 	if ((se = matchchar(s, c, '}')) == NULL) {
@@ -594,17 +564,14 @@ doDefine(MacroBuf mb, const char * se, int level, int expandbody)
 	goto exit;
     }
 
-    /* Options must be terminated with ')' */
-    if (o && oc != ')') {
-	rpmlog(RPMLOG_ERR, _("Macro %%%s has unterminated opts\n"), n);
-	goto exit;
-    }
-
     if ((be - b) < 1) {
 	rpmlog(RPMLOG_ERR, _("Macro %%%s has empty body\n"), n);
 	goto exit;
     }
 
+    if (!isblank(*sbody) && !(*sbody == '\\' && iseol(sbody[1])))
+	rpmlog(RPMLOG_WARNING, _("Macro %%%s needs whitespace before body\n"), n);
+
     if (expandbody) {
 	if (expandThis(mb, b, 0, &ebody)) {
 	    rpmlog(RPMLOG_ERR, _("Macro %%%s failed to expand\n"), n);
@@ -613,7 +580,7 @@ doDefine(MacroBuf mb, const char * se, int level, int expandbody)
 	b = ebody;
     }
 
-    addMacro(mb->mc, n, o, b, (level - 1));
+    pushMacro(mb->mc, n, o, b, (level - 1), ME_NONE);
 
 exit:
     _free(buf);
@@ -625,13 +592,14 @@ exit:
  * Parse (and execute) macro undefinition.
  * @param mc		macro context
  * @param se		macro name to undefine
+ * @param slen		length of se argument
  * @return		address to continue parsing
  */
 static const char *
-doUndefine(rpmMacroContext mc, const char * se)
+doUndefine(rpmMacroContext mc, const char * se, size_t slen)
 {
     const char *s = se;
-    char *buf = xmalloc(MACROBUFSIZ);
+    char *buf = xmalloc(slen + 1);
     char *n = buf, *ne = n;
     int c;
 
@@ -649,103 +617,44 @@ doUndefine(rpmMacroContext mc, const char * se)
 	goto exit;
     }
 
-    delMacro(mc, n);
+    popMacro(mc, n);
 
 exit:
     _free(buf);
     return se;
 }
 
-/**
- * Push new macro definition onto macro entry stack.
- * @param mep		address of macro entry slot
- * @param n		macro name
- * @param o		macro parameters (NULL if none)
- * @param b		macro body (NULL becomes "")
- * @param level		macro recursion level
- */
-static void
-pushMacro(rpmMacroEntry * mep,
-		const char * n, const char * o,
-		const char * b, int level)
-{
-    rpmMacroEntry prev = (mep && *mep ? *mep : NULL);
-    rpmMacroEntry me = (rpmMacroEntry) xmalloc(sizeof(*me));
-
-    me->prev = prev;
-    me->name = (prev ? prev->name : xstrdup(n));
-    me->opts = (o ? xstrdup(o) : NULL);
-    me->body = xstrdup(b ? b : "");
-    me->used = 0;
-    me->level = level;
-    if (mep)
-	*mep = me;
-    else
-	me = _free(me);
-}
-
-/**
- * Pop macro definition from macro entry stack.
- * @param mep		address of macro entry slot
- */
-static void
-popMacro(rpmMacroEntry * mep)
-{
-	rpmMacroEntry me = (*mep ? *mep : NULL);
-
-	if (me) {
-		/* XXX cast to workaround const */
-		if ((*mep = me->prev) == NULL)
-			me->name = _free(me->name);
-		me->opts = _free(me->opts);
-		me->body = _free(me->body);
-		me = _free(me);
-	}
-}
-
 /**
  * Free parsed arguments for parameterized macro.
  * @param mb		macro expansion state
  */
 static void
-freeArgs(MacroBuf mb)
+freeArgs(MacroBuf mb, int delete)
 {
     rpmMacroContext mc = mb->mc;
-    int ndeleted = 0;
-    int i;
-
-    if (mc == NULL || mc->macroTable == NULL)
-	return;
 
     /* Delete dynamic macro definitions */
-    for (i = 0; i < mc->firstFree; i++) {
-	rpmMacroEntry *mep, me;
-	int skiptest = 0;
-	mep = &mc->macroTable[i];
-	me = *mep;
-
-	if (me == NULL)		/* XXX this should never happen */
-	    continue;
+    for (int i = 0; i < mc->n; i++) {
+	rpmMacroEntry me = mc->tab[i];
+	assert(me);
 	if (me->level < mb->depth)
 	    continue;
-	if (strlen(me->name) == 1 && strchr("#*0", *me->name)) {
-	    if (*me->name == '*' && me->used > 0)
-		skiptest = 1; /* XXX skip test for %# %* %0 */
-	} else if (!skiptest && me->used <= 0) {
-#if NOTYET
-	    rpmlog(RPMLOG_ERR,
-			_("Macro %%%s (%s) was not used below level %d\n"),
-			me->name, me->body, me->level);
-#endif
+	/* Warn on defined but unused non-automatic, scoped macros */
+	if (!(me->flags & (ME_AUTO|ME_USED))) {
+	    rpmlog(RPMLOG_WARNING,
+			_("Macro %%%s defined but not used within scope\n"),
+			me->name);
+	    /* Only whine once */
+	    me->flags |= ME_USED;
 	}
-	popMacro(mep);
-	if (!(mep && *mep))
-	    ndeleted++;
-    }
+	if (!delete)
+	    continue;
 
-    /* If any deleted macros, sort macro table */
-    if (ndeleted)
-	sortMacroTable(mc);
+	/* compensate if the slot is to go away */
+	if (me->prev == NULL)
+	    i--;
+	popMacro(mc, me->name);
+    }
 }
 
 /**
@@ -769,7 +678,7 @@ grabArgs(MacroBuf mb, const rpmMacroEntry me, const char * se,
 
     /* Copy macro name as argv[0] */
     argvAdd(&argv, me->name);
-    addMacro(mb->mc, "0", NULL, me->name, mb->depth);
+    pushMacro(mb->mc, "0", NULL, me->name, mb->depth, ME_AUTO);
     
     /* 
      * Make a copy of se up to lastc string that we can pass to argvSplit().
@@ -795,7 +704,7 @@ grabArgs(MacroBuf mb, const rpmMacroEntry me, const char * se,
      * This is the (potential) justification for %{**} ...
     */
     args = argvJoin(argv + 1, " ");
-    addMacro(mb->mc, "**", NULL, args, mb->depth);
+    pushMacro(mb->mc, "**", NULL, args, mb->depth, ME_AUTO);
     free(args);
 
     /*
@@ -817,7 +726,7 @@ grabArgs(MacroBuf mb, const rpmMacroEntry me, const char * se,
 	char *name = NULL, *body = NULL;
 	if (c == '?' || strchr(opts, c) == NULL) {
 	    rpmlog(RPMLOG_ERR, _("Unknown option %c in %s(%s)\n"),
-			(char)c, me->name, opts);
+			(char)optopt, me->name, opts);
 	    goto exit;
 	}
 
@@ -827,13 +736,13 @@ grabArgs(MacroBuf mb, const rpmMacroEntry me, const char * se,
 	} else {
 	    rasprintf(&body, "-%c", c);
 	}
-	addMacro(mb->mc, name, NULL, body, mb->depth);
+	pushMacro(mb->mc, name, NULL, body, mb->depth, ME_AUTO);
 	free(name);
 	free(body);
 
 	if (optarg) {
 	    rasprintf(&name, "-%c*", c);
-	    addMacro(mb->mc, name, NULL, optarg, mb->depth);
+	    pushMacro(mb->mc, name, NULL, optarg, mb->depth, ME_AUTO);
 	    free(name);
 	}
     }
@@ -841,7 +750,7 @@ grabArgs(MacroBuf mb, const rpmMacroEntry me, const char * se,
     /* Add argument count (remaining non-option items) as macro. */
     {	char *ac = NULL;
     	rasprintf(&ac, "%d", (argc - optind));
-    	addMacro(mb->mc, "#", NULL, ac, mb->depth);
+    	pushMacro(mb->mc, "#", NULL, ac, mb->depth, ME_AUTO);
 	free(ac);
     }
 
@@ -850,19 +759,20 @@ grabArgs(MacroBuf mb, const rpmMacroEntry me, const char * se,
 	for (c = optind; c < argc; c++) {
 	    char *name = NULL;
 	    rasprintf(&name, "%d", (c - optind + 1));
-	    addMacro(mb->mc, name, NULL, argv[c], mb->depth);
+	    pushMacro(mb->mc, name, NULL, argv[c], mb->depth, ME_AUTO);
 	    free(name);
 	}
     }
 
     /* Add concatenated unexpanded arguments as yet another macro. */
     args = argvJoin(argv + optind, " ");
-    addMacro(mb->mc, "*", NULL, args ? args : "", mb->depth);
+    pushMacro(mb->mc, "*", NULL, args ? args : "", mb->depth, ME_AUTO);
     free(args);
 
 exit:
     argvFree(argv);
-    return *lastc ? lastc + 1 : lastc; 
+    return ((*lastc == '\0' || *lastc == '\n') && *(lastc-1) != '\\') ?
+	   lastc : lastc + 1;
 }
 
 /**
@@ -902,7 +812,7 @@ doFoo(MacroBuf mb, int negate, const char * f, size_t fn,
     char *b = NULL, *be;
     int c;
 
-    if (g != NULL) {
+    if (g != NULL && gn > 0) {
 	(void) expandThis(mb, g, gn, &buf);
     } else {
 	buf = xmalloc(MACROBUFSIZ + fn + gn);
@@ -913,13 +823,10 @@ doFoo(MacroBuf mb, int negate, const char * f, size_t fn,
 	    b = buf;
 	else
 	    b++;
-#if NOTYET
-    /* XXX watchout for conflict with %dir */
     } else if (STREQ("dirname", f, fn)) {
 	if ((b = strrchr(buf, '/')) != NULL)
 	    *b = '\0';
 	b = buf;
-#endif
     } else if (STREQ("suffix", f, fn)) {
 	if ((b = strrchr(buf, '.')) != NULL)
 	    b++;
@@ -965,6 +872,9 @@ doFoo(MacroBuf mb, int negate, const char * f, size_t fn,
 	case COMPRESSED_LRZIP:
 	    sprintf(be, "%%__lrzip -dqo- %s", b);
 	    break;
+	case COMPRESSED_7ZIP:
+	    sprintf(be, "%%__7zip x %s", b);
+	    break;
 	}
 	b = be;
     } else if (STREQ("getenv", f, fn)) {
@@ -1021,7 +931,12 @@ expandMacro(MacroBuf mb, const char *src, size_t slen)
     int chkexist;
     char *source = NULL;
 
-    /* Handle non-terminated substrings by creating a terminated copy */
+    /*
+     * Always make a (terminated) copy of the source string.
+     * Beware: avoiding the copy when src is known \0-terminated seems like
+     * an obvious opportunity for optimization, but doing that breaks
+     * a special case of macro undefining itself.
+     */
     if (!slen)
 	slen = strlen(src);
     source = xmalloc(slen + 1);
@@ -1030,8 +945,9 @@ expandMacro(MacroBuf mb, const char *src, size_t slen)
     s = source;
 
     if (mb->buf == NULL) {
-	size_t blen = MACROBUFSIZ + strlen(s);
-	mb->buf = xcalloc(blen + 1, sizeof(*mb->buf));
+	size_t blen = MACROBUFSIZ + slen;
+	mb->buf = xmalloc(blen + 1);
+	mb->buf[0] = '\0';
 	mb->tpos = 0;
 	mb->nb = blen;
     }
@@ -1176,16 +1092,29 @@ expandMacro(MacroBuf mb, const char *src, size_t slen)
 		printMacro(mb, s, se);
 
 	/* Expand builtin macros */
+	if (STREQ("load", f, fn)) {
+		char *arg = NULL;
+		if (g && gn > 0 && expandThis(mb, g, gn, &arg) == 0) {
+			/* Print failure iff %{load:...} or %{!?load:...} */
+			if (loadMacroFile(mb->mc, arg) && chkexist == negate) {
+				rpmlog(RPMLOG_ERR,
+				       _("failed to load macro file %s"), arg);
+			}
+		}
+		free(arg);
+		s = se;
+		continue;
+	}
 	if (STREQ("global", f, fn)) {
-		s = doDefine(mb, se, RMIL_GLOBAL, 1);
+		s = doDefine(mb, se, slen - (se - s), RMIL_GLOBAL, 1);
 		continue;
 	}
 	if (STREQ("define", f, fn)) {
-		s = doDefine(mb, se, mb->depth, 0);
+		s = doDefine(mb, se, slen - (se - s), mb->depth, 0);
 		continue;
 	}
 	if (STREQ("undefine", f, fn)) {
-		s = doUndefine(mb->mc, se);
+		s = doUndefine(mb->mc, se, slen - (se - s));
 		continue;
 	}
 
@@ -1225,12 +1154,11 @@ expandMacro(MacroBuf mb, const char *src, size_t slen)
 #ifdef	WITH_LUA
 	if (STREQ("lua", f, fn)) {
 		rpmlua lua = NULL; /* Global state. */
-		const char *ls = s+sizeof("{lua:")-1;
-		const char *lse = se-sizeof("}")+1;
-		char *scriptbuf = (char *)xmalloc((lse-ls)+1);
+		char *scriptbuf = xmalloc(gn + 1);
 		char *printbuf;
-		memcpy(scriptbuf, ls, lse-ls);
-		scriptbuf[lse-ls] = '\0';
+		if (g != NULL && gn > 0) 
+		    memcpy(scriptbuf, g, gn);
+		scriptbuf[gn] = '\0';
 		rpmluaPushPrintBuffer(lua);
 		if (rpmluaRunScript(lua, scriptbuf, NULL) == -1)
 		    rc = 1;
@@ -1247,6 +1175,7 @@ expandMacro(MacroBuf mb, const char *src, size_t slen)
 
 	/* XXX necessary but clunky */
 	if (STREQ("basename", f, fn) ||
+	    STREQ("dirname", f, fn) ||
 	    STREQ("suffix", f, fn) ||
 	    STREQ("expand", f, fn) ||
 	    STREQ("verbose", f, fn) ||
@@ -1265,13 +1194,15 @@ expandMacro(MacroBuf mb, const char *src, size_t slen)
 	}
 
 	/* Expand defined macros */
-	mep = findEntry(mb->mc, f, fn);
+	mep = findEntry(mb->mc, f, fn, NULL);
 	me = (mep ? *mep : NULL);
 
+	/* If we looked up a macro, consider it used */
+	if (me)
+	    me->flags |= ME_USED;
+
 	/* XXX Special processing for flags */
 	if (*f == '-') {
-		if (me)
-			me->used++;	/* Mark macro as used */
 		if ((me == NULL && !negate) ||	/* Without -f, skip %{-f...} */
 		    (me != NULL && negate)) {	/* With -f, skip %{!-f...} */
 			s = se;
@@ -1317,28 +1248,29 @@ expandMacro(MacroBuf mb, const char *src, size_t slen)
 		if (lastc != NULL) {
 			se = grabArgs(mb, me, fe, lastc);
 		} else {
-			addMacro(mb->mc, "**", NULL, "", mb->depth);
-			addMacro(mb->mc, "*", NULL, "", mb->depth);
-			addMacro(mb->mc, "#", NULL, "0", mb->depth);
-			addMacro(mb->mc, "0", NULL, me->name, mb->depth);
+			pushMacro(mb->mc, "**", NULL, "", mb->depth, ME_AUTO);
+			pushMacro(mb->mc, "*", NULL, "", mb->depth, ME_AUTO);
+			pushMacro(mb->mc, "#", NULL, "0", mb->depth, ME_AUTO);
+			pushMacro(mb->mc, "0", NULL, me->name, mb->depth, ME_AUTO);
 		}
 	}
 
 	/* Recursively expand body of macro */
 	if (me->body && *me->body) {
 		rc = expandMacro(mb, me->body, 0);
-		if (rc == 0)
-			me->used++;	/* Mark macro as used */
 	}
 
 	/* Free args for "%name " macros with opts */
 	if (me->opts != NULL)
-		freeArgs(mb);
+		freeArgs(mb, 1);
 
 	s = se;
     }
 
     mb->buf[mb->tpos] = '\0';
+    /* Warn on unused macros, but only on error/tracing as its very expesive */
+    if (rc != 0 || mb->expand_trace)
+	freeArgs(mb, 0);
     mb->depth--;
     if (rc != 0 || mb->expand_trace)
 	printExpansion(mb, mb->buf+tpos, mb->buf+mb->tpos);
@@ -1354,8 +1286,6 @@ static int doExpandMacros(rpmMacroContext mc, const char *src, char **target)
     MacroBuf mb = xcalloc(1, sizeof(*mb));
     int rc = 0;
 
-    if (mc == NULL) mc = rpmGlobalMacroContext;
-
     mb->buf = NULL;
     mb->depth = 0;
     mb->macro_trace = print_macro_trace;
@@ -1372,101 +1302,117 @@ static int doExpandMacros(rpmMacroContext mc, const char *src, char **target)
     return rc;
 }
 
-int expandMacros(void * spec, rpmMacroContext mc, char * sbuf, size_t slen)
+static void pushMacro(rpmMacroContext mc,
+	const char * n, const char * o, const char * b, int level, int flags)
 {
-    char *target = NULL;
-    int rc = doExpandMacros(mc, sbuf, &target);
-    rstrlcpy(sbuf, target, slen);
-    free(target);
-    return rc;
-}
+    /* new entry */
+    rpmMacroEntry me;
+    /* pointer into me */
+    char *p;
+    /* calculate sizes */
+    size_t olen = o ? strlen(o) : 0;
+    size_t blen = b ? strlen(b) : 0;
+    size_t mesize = sizeof(*me) + blen + 1 + (olen ? olen + 1 : 0);
 
-void
-addMacro(rpmMacroContext mc,
-	const char * n, const char * o, const char * b, int level)
-{
-    rpmMacroEntry * mep;
-
-    if (mc == NULL) mc = rpmGlobalMacroContext;
-
-    /* If new name, expand macro table */
-    if ((mep = findEntry(mc, n, 0)) == NULL) {
-	if (mc->firstFree == mc->macrosAllocated)
-	    expandMacroTable(mc);
-	if (mc->macroTable != NULL)
-	    mep = mc->macroTable + mc->firstFree++;
+    size_t pos;
+    rpmMacroEntry *mep = findEntry(mc, n, 0, &pos);
+    if (mep) {
+	/* entry with shared name */
+	me = xmalloc(mesize);
+	/* copy body */
+	me->body = p = me->arena;
+	if (blen)
+	    memcpy(p, b, blen + 1);
+	else
+	    *p = '\0';
+	p += blen + 1;
+	/* set name */
+	me->name = (*mep)->name;
+    }
+    else {
+	/* extend macro table */
+	const int delta = 256;
+	if (mc->n % delta == 0)
+	    mc->tab = xrealloc(mc->tab, sizeof(me) * (mc->n + delta));
+	/* shift pos+ entries to the right */
+	memmove(mc->tab + pos + 1, mc->tab + pos, sizeof(me) * (mc->n - pos));
+	mc->n++;
+	/* make slot */
+	mc->tab[pos] = NULL;
+	mep = &mc->tab[pos];
+	/* entry with new name */
+	size_t nlen = strlen(n);
+	me = xmalloc(mesize + nlen + 1);
+	/* copy body */
+	me->body = p = me->arena;
+	if (blen)
+	    memcpy(p, b, blen + 1);
+	else
+	    *p = '\0';
+	p += blen + 1;
+	/* copy name */
+	me->name = memcpy(p, n, nlen + 1);
+	p += nlen + 1;
     }
 
-    if (mep != NULL) {
-	/* Push macro over previous definition */
-	pushMacro(mep, n, o, b, level);
-
-	/* If new name, sort macro table */
-	if ((*mep)->prev == NULL)
-	    sortMacroTable(mc);
-    }
+    /* copy options */
+    if (olen)
+	me->opts = memcpy(p, o, olen + 1);
+    else
+	me->opts = o ? "" : NULL;
+    /* initialize */
+    me->flags = flags;
+    me->flags &= ~(ME_USED);
+    me->level = level;
+    /* push over previous definition */
+    me->prev = *mep;
+    *mep = me;
 }
 
-void
-delMacro(rpmMacroContext mc, const char * n)
+static void popMacro(rpmMacroContext mc, const char * n)
 {
-    rpmMacroEntry * mep;
-
-    if (mc == NULL) mc = rpmGlobalMacroContext;
-    /* If name exists, pop entry */
-    if ((mep = findEntry(mc, n, 0)) != NULL) {
-	popMacro(mep);
-	/* If deleted name, sort macro table */
-	if (!(mep && *mep))
-	    sortMacroTable(mc);
+    size_t pos;
+    rpmMacroEntry *mep = findEntry(mc, n, 0, &pos);
+    if (mep == NULL)
+	return;
+    /* parting entry */
+    rpmMacroEntry me = *mep;
+    assert(me);
+    /* detach/pop definition */
+    mc->tab[pos] = me->prev;
+    /* shrink macro table */
+    if (me->prev == NULL) {
+	mc->n--;
+	/* move pos+ elements to the left */
+	memmove(mc->tab + pos, mc->tab + pos + 1, sizeof(me) * (mc->n - pos));
+	/* deallocate */
+	if (mc->n == 0)
+	    mc->tab = _free(mc->tab);
     }
+    /* comes in a single chunk */
+    free(me);
 }
 
-int
-rpmDefineMacro(rpmMacroContext mc, const char * macro, int level)
+static int defineMacro(rpmMacroContext mc, const char * macro, int level)
 {
     MacroBuf mb = xcalloc(1, sizeof(*mb));
 
     /* XXX just enough to get by */
-    mb->mc = (mc ? mc : rpmGlobalMacroContext);
-    (void) doDefine(mb, macro, level, 0);
+    mb->mc = mc;
+    (void) doDefine(mb, macro, strlen(macro), level, 0);
     _free(mb);
     return 0;
 }
 
-void
-rpmLoadMacros(rpmMacroContext mc, int level)
-{
-
-    if (mc == NULL || mc == rpmGlobalMacroContext)
-	return;
-
-    if (mc->macroTable != NULL) {
-	int i;
-	for (i = 0; i < mc->firstFree; i++) {
-	    rpmMacroEntry *mep, me;
-	    mep = &mc->macroTable[i];
-	    me = *mep;
-
-	    if (me == NULL)		/* XXX this should never happen */
-		continue;
-	    addMacro(NULL, me->name, me->opts, me->body, (level - 1));
-	}
-    }
-}
-
-int
-rpmLoadMacroFile(rpmMacroContext mc, const char * fn)
+static int loadMacroFile(rpmMacroContext mc, const char * fn)
 {
     FILE *fd = fopen(fn, "r");
     size_t blen = MACROBUFSIZ;
     char *buf = xmalloc(blen);
     int rc = -1;
 
-    if (fd == NULL || ferror(fd)) {
-	if (fd) (void) fclose(fd);
+    if (fd == NULL)
 	goto exit;
-    }
 
     /* XXX Assume new fangled macro expansion */
     max_macro_depth = 16;
@@ -1481,8 +1427,9 @@ rpmLoadMacroFile(rpmMacroContext mc, const char * fn)
 	if (c != '%')
 		continue;
 	n++;	/* skip % */
-	rc = rpmDefineMacro(mc, n, RMIL_MACROFILES);
+	rc = defineMacro(mc, n, RMIL_MACROFILES);
     }
+
     rc = fclose(fd);
 
 exit:
@@ -1490,15 +1437,117 @@ exit:
     return rc;
 }
 
+static void copyMacros(rpmMacroContext src, rpmMacroContext dst, int level)
+{
+    for (int i = 0; i < src->n; i++) {
+	rpmMacroEntry me = src->tab[i];
+	assert(me);
+	pushMacro(dst, me->name, me->opts, me->body, (level - 1), me->flags);
+    }
+}
+
+/* External interfaces */
+
+int expandMacros(void * spec, rpmMacroContext mc, char * sbuf, size_t slen)
+{
+    char *target = NULL;
+    int rc;
+
+    mc = rpmmctxAcquire(mc);
+    rc = doExpandMacros(mc, sbuf, &target);
+    rpmmctxRelease(mc);
+
+    rstrlcpy(sbuf, target, slen);
+    free(target);
+    return rc;
+}
+
+void
+rpmDumpMacroTable(rpmMacroContext mc, FILE * fp)
+{
+    mc = rpmmctxAcquire(mc);
+    if (fp == NULL) fp = stderr;
+    
+    fprintf(fp, "========================\n");
+    for (int i = 0; i < mc->n; i++) {
+	rpmMacroEntry me = mc->tab[i];
+	assert(me);
+	fprintf(fp, "%3d%c %s", me->level,
+		    ((me->flags & ME_USED) ? '=' : ':'), me->name);
+	if (me->opts && *me->opts)
+		fprintf(fp, "(%s)", me->opts);
+	if (me->body && *me->body)
+		fprintf(fp, "\t%s", me->body);
+	fprintf(fp, "\n");
+    }
+    fprintf(fp, _("======================== active %d empty %d\n"),
+		mc->n, 0);
+    rpmmctxRelease(mc);
+}
+
+void addMacro(rpmMacroContext mc,
+	      const char * n, const char * o, const char * b, int level)
+{
+    mc = rpmmctxAcquire(mc);
+    pushMacro(mc, n, o, b, level, ME_NONE);
+    rpmmctxRelease(mc);
+}
+
+void delMacro(rpmMacroContext mc, const char * n)
+{
+    mc = rpmmctxAcquire(mc);
+    popMacro(mc, n);
+    rpmmctxRelease(mc);
+}
+
+int
+rpmDefineMacro(rpmMacroContext mc, const char * macro, int level)
+{
+    mc = rpmmctxAcquire(mc);
+    (void) defineMacro(mc, macro, level);
+    rpmmctxRelease(mc);
+    return 0;
+}
+
+void
+rpmLoadMacros(rpmMacroContext mc, int level)
+{
+    rpmMacroContext gmc;
+    if (mc == NULL || mc == rpmGlobalMacroContext)
+	return;
+
+    gmc = rpmmctxAcquire(NULL);
+    mc = rpmmctxAcquire(mc);
+
+    copyMacros(mc, gmc, level);
+
+    rpmmctxRelease(mc);
+    rpmmctxRelease(gmc);
+}
+
+int
+rpmLoadMacroFile(rpmMacroContext mc, const char * fn)
+{
+    int rc;
+
+    mc = rpmmctxAcquire(mc);
+    rc = loadMacroFile(mc, fn);
+    rpmmctxRelease(mc);
+
+    return rc;
+}
+
 void
 rpmInitMacros(rpmMacroContext mc, const char * macrofiles)
 {
     ARGV_t pattern, globs = NULL;
+    rpmMacroContext climc;
 
     if (macrofiles == NULL)
 	return;
 
     argvSplit(&globs, macrofiles, ":");
+    mc = rpmmctxAcquire(mc);
     for (pattern = globs; *pattern; pattern++) {
 	ARGV_t path, files = NULL;
     
@@ -1514,38 +1563,30 @@ rpmInitMacros(rpmMacroContext mc, const char * macrofiles)
 		rpmFileHasSuffix(*path, ".rpmorig")) {
 		continue;
 	    }
-	    (void) rpmLoadMacroFile(mc, *path);
+	    (void) loadMacroFile(mc, *path);
 	}
 	argvFree(files);
     }
     argvFree(globs);
 
     /* Reload cmdline macros */
-    rpmLoadMacros(rpmCLIMacroContext, RMIL_CMDLINE);
+    climc = rpmmctxAcquire(rpmCLIMacroContext);
+    copyMacros(climc, mc, RMIL_CMDLINE);
+    rpmmctxRelease(climc);
+
+    rpmmctxRelease(mc);
 }
 
 void
 rpmFreeMacros(rpmMacroContext mc)
 {
-    
-    if (mc == NULL) mc = rpmGlobalMacroContext;
-
-    if (mc->macroTable != NULL) {
-	int i;
-	for (i = 0; i < mc->firstFree; i++) {
-	    rpmMacroEntry me;
-	    while ((me = mc->macroTable[i]) != NULL) {
-		/* XXX cast to workaround const */
-		if ((mc->macroTable[i] = me->prev) == NULL)
-		    me->name = _free(me->name);
-		me->opts = _free(me->opts);
-		me->body = _free(me->body);
-		me = _free(me);
-	    }
-	}
-	mc->macroTable = _free(mc->macroTable);
+    mc = rpmmctxAcquire(mc);
+    while (mc->n > 0) {
+	/* remove from the end to avoid memmove */
+	rpmMacroEntry me = mc->tab[mc->n - 1];
+	popMacro(mc, me->name);
     }
-    memset(mc, 0, sizeof(*mc));
+    rpmmctxRelease(mc);
 }
 
 char * 
@@ -1556,6 +1597,7 @@ rpmExpand(const char *arg, ...)
     char *pe;
     const char *s;
     va_list ap;
+    rpmMacroContext mc;
 
     if (arg == NULL) {
 	ret = xstrdup("");
@@ -1576,7 +1618,9 @@ rpmExpand(const char *arg, ...)
 	pe = stpcpy(pe, s);
     va_end(ap);
 
-    (void) doExpandMacros(NULL, buf, &ret);
+    mc = rpmmctxAcquire(NULL);
+    (void) doExpandMacros(mc, buf, &ret);
+    rpmmctxRelease(mc);
 
     free(buf);
 exit:
@@ -1605,7 +1649,7 @@ rpmExpandNumeric(const char *arg)
 	if (!(end && *end == '\0'))
 	    rc = 0;
     }
-    val = _free(val);
+    free(val);
 
     return rc;
 }
