@@ -7,6 +7,26 @@
 #include <lposix.h>
 #include <lrexlib.h>
 
+/* replaced in 5.1 */
+#ifndef lua_open
+#define lua_open()	luaL_newstate()
+#endif
+
+/* defined as lua_objlen in 5.1 */
+#ifndef lua_strlen
+#define lua_strlen(L,i)	lua_rawlen(L, (i))
+#endif
+
+/* deprecated in 5.1, defined as lua_objlen in 5.1 */
+#ifndef luaL_getn
+#define luaL_getn(L,i)	((int)lua_rawlen(L, i))
+#endif
+
+/* define added in 5.2 */
+#ifndef lua_pushglobaltable
+#define lua_pushglobaltable(L)	lua_pushvalue(L, LUA_GLOBALSINDEX)
+#endif
+
 #include <unistd.h>
 #include <assert.h>
 
@@ -15,8 +35,8 @@
 #include <rpm/rpmlog.h>
 #include <rpm/rpmurl.h>
 #include <rpm/rpmfileutil.h>
+#include <rpm/rpmbase64.h>
 #include "rpmio/rpmhook.h"
-#include "rpmio/base64.h"
 
 #define _RPMLUA_INTERNAL
 #include "rpmio/rpmlua.h"
@@ -53,10 +73,10 @@ rpmlua rpmluaNew()
 {
     rpmlua lua = (rpmlua) xcalloc(1, sizeof(*lua));
     struct stat st;
-    const luaL_reg *lib;
+    const luaL_Reg *lib;
     char *initlua = rpmGenPath(rpmConfigDir(), "init.lua", NULL);
    
-    static const luaL_reg extlibs[] = {
+    static const luaL_Reg extlibs[] = {
 	{"posix", luaopen_posix},
 	{"rex", luaopen_rex},
 	{"rpm", luaopen_rpm},
@@ -74,12 +94,26 @@ rpmlua rpmluaNew()
 	lua_call(L, 1, 0);
 	lua_settop(L, 0);
     }
+#ifndef LUA_GLOBALSINDEX
+    lua_pushglobaltable(L);
+#endif
     lua_pushliteral(L, "LUA_PATH");
     lua_pushfstring(L, "%s/%s", rpmConfigDir(), "/lua/?.lua");
+#ifdef LUA_GLOBALSINDEX
     lua_rawset(L, LUA_GLOBALSINDEX);
+#else
+    lua_settable(L, -3);
+#endif
     lua_pushliteral(L, "print");
     lua_pushcfunction(L, rpm_print);
+#ifdef LUA_GLOBALSINDEX
     lua_rawset(L, LUA_GLOBALSINDEX);
+#else
+    lua_settable(L, -3);
+#endif
+#ifndef LUA_GLOBALSINDEX
+    lua_pop(L, 1);
+#endif
     rpmluaSetData(lua, "lua", lua);
     if (stat(initlua, &st) != -1)
 	(void)rpmluaRunScriptFile(lua, initlua);
@@ -191,7 +225,7 @@ void rpmluaSetVar(rpmlua _lua, rpmluav var)
     }
     if (!var->listmode || lua->pushsize > 0) {
 	if (lua->pushsize == 0)
-	    lua_pushvalue(L, LUA_GLOBALSINDEX);
+	    lua_pushglobaltable(L);
 	if (pushvar(L, var->keyType, &var->key) != -1) {
 	    if (pushvar(L, var->valueType, &var->value) != -1)
 		lua_rawset(L, -3);
@@ -228,7 +262,7 @@ void rpmluaGetVar(rpmlua _lua, rpmluav var)
     lua_State *L = lua->L;
     if (!var->listmode) {
 	if (lua->pushsize == 0)
-	    lua_pushvalue(L, LUA_GLOBALSINDEX);
+	    lua_pushglobaltable(L);
 	if (pushvar(L, var->keyType, &var->key) != -1) {
 	    lua_rawget(L, -2);
 	    popvar(L, &var->valueType, &var->value);
@@ -261,7 +295,7 @@ static int findkey(lua_State *L, int oper, const char *key, va_list va)
     vsnprintf(buf, blen + 1, key, va);
 
     s = e = buf;
-    lua_pushvalue(L, LUA_GLOBALSINDEX);
+    lua_pushglobaltable(L);
     for (;;) {
 	if (*e == '\0' || *e == '.') {
 	    if (e != s) {
@@ -492,6 +526,8 @@ int rpmluaRunScript(rpmlua _lua, const char *script, const char *name)
     int ret = 0;
     if (name == NULL)
 	name = "<lua>";
+    if (script == NULL)
+	script = "";
     if (luaL_loadbuffer(L, script, strlen(script), name) != 0) {
 	rpmlog(RPMLOG_ERR, _("invalid syntax in lua script: %s\n"),
 		 lua_tostring(L, -1));
@@ -596,7 +632,7 @@ static int rpm_b64encode(lua_State *L)
     if (lua_gettop(L) == 2)
 	linelen = luaL_checkinteger(L, 2);
     if (str && len) {
-	char *data = b64encode(str, len, linelen);
+	char *data = rpmBase64Encode(str, len, linelen);
 	lua_pushstring(L, data);
 	free(data);
     }
@@ -609,7 +645,7 @@ static int rpm_b64decode(lua_State *L)
     if (str) {
 	void *data = NULL;
 	size_t len = 0;
-	if (b64decode(str, &data, &len) == 0) {
+	if (rpmBase64Decode(str, &data, &len) == 0) {
 	    lua_pushlstring(L, data, len);
 	} else {
 	    lua_pushnil(L);
@@ -635,6 +671,14 @@ static int rpm_define(lua_State *L)
     return 0;
 }
 
+static int rpm_load(lua_State *L)
+{
+    const char *str = luaL_checkstring(L, 1);
+    int rc = rpmLoadMacroFile(NULL, str);
+    lua_pushnumber(L, rc);
+    return 1;
+}
+
 static int rpm_interactive(lua_State *L)
 {
     _rpmluaInteractive(L);
@@ -822,11 +866,12 @@ static int rpm_print (lua_State *L)
     return 0;
 }
 
-static const luaL_reg rpmlib[] = {
+static const luaL_Reg rpmlib[] = {
     {"b64encode", rpm_b64encode},
     {"b64decode", rpm_b64decode},
     {"expand", rpm_expand},
     {"define", rpm_define},
+    {"load", rpm_load},
     {"register", rpm_register},
     {"unregister", rpm_unregister},
     {"call", rpm_call},
@@ -836,7 +881,7 @@ static const luaL_reg rpmlib[] = {
 
 static int luaopen_rpm(lua_State *L)
 {
-    lua_pushvalue(L, LUA_GLOBALSINDEX);
+    lua_pushglobaltable(L);
     luaL_openlib(L, "rpm", rpmlib, 0);
     return 0;
 }
