@@ -11,16 +11,26 @@ test -x $pkgconfig || {
     exit 0
 }
 
+$pkgconfig --atleast-pkgconfig-version="0.24" || {
+    cat > /dev/null
+    exit 0
+}
+
 case $1 in
 -P|--provides)
     while read filename ; do
     case "${filename}" in
     *.pc)
 	# Query the dependencies of the package.
+	DIR="`dirname ${filename}`"
+	export PKG_CONFIG_PATH="$DIR:$DIR/../../share/pkgconfig"
 	$pkgconfig --print-provides "$filename" 2> /dev/null | while read n r v ; do
+	    [ -n "$n" ] || continue
 	    # We have a dependency.  Make a note that we need the pkgconfig
 	    # tool for this package.
-	    echo "pkgconfig($n)" "$r" "$v"
+	    echo -n "pkgconfig($n) "
+	    [ -n "$r" ] && [ -n "$v" ] && echo -n "$r" "$v"
+	    echo
 	done
 	;;
     esac
@@ -30,10 +40,15 @@ case $1 in
     while read filename ; do
     case "${filename}" in
     *.pc)
-	$pkgconfig --print-requires "$filename" 2> /dev/null | while read n r v ; do
-	    i="`expr $i + 1`"
-	    [ $i -eq 1 ] && echo "pkgconfig"
-	    echo "pkgconfig($n)" "$r" "$v"
+	i="`expr $i + 1`"
+	[ $i -eq 1 ] && echo "$pkgconfig"
+	DIR="`dirname ${filename}`"
+	export PKG_CONFIG_PATH="$DIR:$DIR/../../share/pkgconfig"
+	$pkgconfig --print-requires --print-requires-private "$filename" 2> /dev/null | while read n r v ; do
+	    [ -n "$n" ] || continue
+	    echo -n "pkgconfig($n) "
+	    [ -n "$r" ] && [ -n "$v" ] && echo -n "$r" "$v"
+	    echo
 	done
     esac
     done
