@@ -225,6 +225,8 @@ static struct
 #define DEBUG_FRAME	9
 #define DEBUG_RANGES	10
 #define DEBUG_TYPES	11
+#define DEBUG_MACRO	12
+#define DEBUG_GDB_SCRIPT	13
     { ".debug_info", NULL, NULL, 0, 0, 0 },
     { ".debug_abbrev", NULL, NULL, 0, 0, 0 },
     { ".debug_line", NULL, NULL, 0, 0, 0 },
@@ -237,6 +239,8 @@ static struct
     { ".debug_frame", NULL, NULL, 0, 0, 0 },
     { ".debug_ranges", NULL, NULL, 0, 0, 0 },
     { ".debug_types", NULL, NULL, 0, 0, 0 },
+    { ".debug_macro", NULL, NULL, 0, 0, 0 },
+    { ".debug_gdb_scripts", NULL, NULL, 0, 0, 0 },
     { NULL, NULL, NULL, 0, 0, 0 }
   };
 
@@ -359,7 +363,6 @@ static char *
 canonicalize_path (const char *s, char *d)
 {
   char *rv = d;
-  const char *sroot;
   char *droot;
 
   if (IS_DIR_SEPARATOR (*s))
@@ -375,7 +378,6 @@ canonicalize_path (const char *s, char *d)
 	s++;
     }
   droot = d;
-  sroot = s;
 
   while (*s)
     {
@@ -484,6 +486,10 @@ edit_dwarf2_line (DSO *dso, uint32_t off, char *comp_dir, int phase)
   if (phase != 0)
     return 0;
 
+  /* XXX: RhBug:929365, should we error out instead of ignoring? */
+  if (ptr == NULL)
+    return 0;
+
   ptr += off;
 
   endcu = ptr + 4;
@@ -1154,6 +1160,16 @@ edit_dwarf2 (DSO *dso)
 		  if (rtype != R_ALPHA_REFLONG)
 		    goto fail;
 		  break;
+#if defined(EM_AARCH64) && defined(R_AARCH64_ABS32)
+		case EM_AARCH64:
+		  if (rtype != R_AARCH64_ABS32)
+		    goto fail;
+		  break;
+#endif
+		case EM_68K:
+		  if (rtype != R_68K_32)
+		    goto fail;
+		  break;
 		default:
 		fail:
 		  error (1, 0, "%s: Unhandled relocation %d in .debug_info section",
@@ -1525,6 +1541,12 @@ main (int argc, char *argv[])
 	}
     }
 
+  /* Ensure clean paths, users can muck with these */
+  if (base_dir)
+    canonicalize_path(base_dir, base_dir);
+  if (dest_dir)
+    canonicalize_path(dest_dir, dest_dir);
+
   /* Make sure there are trailing slashes in dirs */
   if (base_dir != NULL && base_dir[strlen (base_dir)-1] != '/')
     {
@@ -1588,7 +1610,7 @@ main (int argc, char *argv[])
 	  if (strcmp (name, ".stab") == 0)
 	    {
 	      fprintf (stderr, "Stabs debuginfo not supported: %s\n", file);
-	      exit (1);
+	      break;
 	    }
 	  if (strcmp (name, ".debug_info") == 0)
 	    edit_dwarf2 (dso);
